from math import floor
import logging
import os

from tqdm.auto import tqdm, trange
import numpy as np
import cupy as cp
from scipy.interpolate import Akima1DInterpolator
from scipy.sparse import coo_matrix

from .postprocess import my_conv2_cpu
from .learn import extractTemplatesfromSnippets
from .utils import get_cuda, Bunch

logger = logging.getLogger(__name__)


def getClosestChannels2(ycup, xcup, yc, xc, NchanClosest):
    # this function outputs the closest channels to each channel,
    # as well as a Gaussian-decaying mask as a function of pairwise distances
    # sigma is the standard deviation of this Gaussian-mask
    # compute distances between all pairs of channels
    xc = cp.asarray(xc, dtype=np.float32, order='F')
    yc = cp.asarray(yc, dtype=np.float32, order='F')
    xcup = cp.asarray(xcup, dtype=np.float32, order='F')
    ycup = cp.asarray(ycup, dtype=np.float32, order='F')
    C2C = ((xc[:, np.newaxis] - xcup[:].T.flatten()[:, np.newaxis].T) ** 2 + (
                yc[:, np.newaxis] - ycup[:].T.flatten()[:, np.newaxis].T) ** 2)
    C2C = cp.sqrt(C2C)
    Nchan, NchanUp = C2C.shape

    # sort distances
    isort = cp.argsort(C2C, axis=0)

    # take NchanCLosest neighbors for each primary channel
    iC = isort[:NchanClosest, :]

    # in some cases we want a mask that decays as a function of distance between pairs of channels
    # this is an awkward indexing to get the corresponding distances
    ix = iC + cp.arange(0, Nchan * NchanUp, Nchan)
    dist = C2C.T.ravel()[ix]

    return iC, dist


def get_batch(params, probe, ibatch, Nbatch, proc) -> cp.ndarray:
    batchstart = np.arange(0, params.NT * Nbatch + 1, params.NT).astype(np.int64)

    offset = probe.Nchan * batchstart[ibatch]
    dat = proc.flat[offset : offset + params.NT * probe.Nchan].reshape(
        (-1, probe.Nchan), order="F"
    )

    # move data to GPU and scale it back to unit variance
    dataRAW = cp.asarray(dat, dtype=np.float32) / params.scaleproc
    return dataRAW


def spikedetector3(Params, drez, wTEMP, iC, dist, v2, iC2, dist2):
    code, constants = get_cuda("spikedetector3")

    NT = int(Params[0])
    Nchan = int(Params[1])
    nt0 = int(Params[2])
    Nnearest = int(Params[3])
    Nrank = int(Params[4])
    NchanUp = int(Params[7])

    Nthreads = constants.Nthreads
    NrankMax = constants.NrankMax
    maxFR = constants.maxFR
    nt0max = constants.nt0max
    NchanMax = constants.NchanMax
    nsizes = constants.nsizes

    # tpB = (8, 2 * nt0 - 1)
    # tpF = (16, Nnearest)
    tpS = (nt0, 16)

    d_Params = cp.asarray(Params, dtype=np.float64, order="F")
    d_data = cp.asarray(drez, dtype=np.float32, order="F")
    d_W = cp.asarray(wTEMP, dtype=np.float32, order="F")
    d_iC = cp.asarray(iC, dtype=np.int32, order="F")
    d_dist = cp.asarray(dist, dtype=np.float32, order="F")
    d_v2 = cp.asarray(v2, dtype=np.float32, order="F")
    d_iC2 = cp.asarray(iC2, dtype=np.int32, order="F")
    d_dist2 = cp.asarray(dist2, dtype=np.float32, order="F")

    dimst = (NT, NchanUp)
    d_dout = cp.zeros(dimst, dtype=np.float32, order="F")
    d_kkmax = cp.zeros(dimst, dtype=np.int32, order="F")

    d_dfilt = cp.zeros((Nrank, NT, Nchan), dtype=np.float32, order="F")
    d_dmax = cp.zeros((NT, NchanUp), dtype=np.float32, order="F")
    d_st = cp.zeros((maxFR * 5), dtype=np.int32, order="F")
    d_cF = cp.zeros((maxFR * Nnearest,), dtype=np.float32, order="F")
    d_counter = cp.zeros(2, dtype=np.int32, order="F")

    # filter the data with the temporal templates
    Conv1D = cp.RawKernel(code, "Conv1D")
    Conv1D((Nchan,), (Nthreads,), (d_Params, d_data, d_W, d_dfilt))

    # sum each template across channels, square, take max
    sumChannels = cp.RawKernel(code, "sumChannels")
    tpP = (int(NT / Nthreads), NchanUp)
    sumChannels(
        tpP, (Nthreads,), (d_Params, d_dfilt, d_dout, d_kkmax, d_iC, d_dist, d_v2)
    )

    # get the max of the data
    max1D = cp.RawKernel(code, "max1D")
    max1D((NchanUp,), (Nthreads,), (d_Params, d_dout, d_dmax))

    # take max across nearby channels
    tpP = (int(NT / Nthreads), NchanUp)
    maxChannels = cp.RawKernel(code, "maxChannels")
    maxChannels(
        tpP,
        (Nthreads,),
        (
            d_Params,
            d_dout,
            d_dmax,
            d_iC,
            d_iC2,
            d_dist2,
            d_kkmax,
            d_dfilt,
            d_st,
            d_counter,
            d_cF,
        ),
    )
    counter = cp.asnumpy(d_counter)[0]

    minSize = min(maxFR, counter)
    d_sto = d_st[: 4 * minSize].reshape((4, minSize), order="F")
    d_cF2 = d_cF[: Nnearest * minSize].reshape((Nnearest, minSize), order="F")

    return d_dout.get(), d_kkmax.get(), d_sto.get(), d_cF2.get()


def kernelD(xp0, yp0, length):
    D = xp0.shape[0]
    N = xp0.shape[1] if len(xp0.shape) > 1 else 1
    M = yp0.shape[1] if len(yp0.shape) > 1 else 1

    K = np.zeros((N, M))
    cs = M

    for i in range(int(M * 1.0 / cs)):
        ii = np.arange(i * cs, min(M, (i + 1) * cs))
        mM = len(ii)

        xp = np.tile(xp0, (mM, 1)).T[np.newaxis, :, :]
        yp = np.tile(yp0[:, ii], (N, 1)).reshape((D, N, mM))
        a = (xp - yp) ** 2
        b = 1.0 / (length ** 2)
        Kn = np.exp(-np.sum((a * b) / 2, axis=0))

        K[:, ii] = Kn

    return K


def align_block2(F, ysamp, nblocks):

    # F is y bins by amp bins by batches
    # ysamp are the coordinates of the y bins in um

    Nbatches = F.shape[2]

    # look up and down this many y bins to find best alignment
    n = 15
    dc = np.zeros((2 * n + 1, Nbatches))
    dt = np.arange(-n, n + 1)

    # we do everything on the GPU for speed, but it's probably fast enough on
    # the CPU
    Fg = F

    # mean subtraction to compute covariance
    Fg = Fg - np.mean(Fg, axis=0)

    # initialize the target "frame" for alignment with a single sample
    F0 = Fg[:, :, min(299, np.floor(Fg.shape[2] / 2).astype("int")) - 1]
    F0 = F0[:, :, np.newaxis]

    # first we do rigid registration by integer shifts
    # everything is iteratively aligned until most of the shifts become 0.
    niter = 10
    dall = np.zeros((niter, Nbatches))
    for iteration in range(niter):
        for t in range(len(dt)):
            # for each NEW potential shift, estimate covariance
            Fs = np.roll(Fg, dt[t], axis=0)
            dc[t, :] = np.mean(Fs * F0, axis=(0,1))
        if iteration + 1 < niter:
            # up until the very last iteration, estimate the best shifts
            imax = np.argmax(dc, axis=0)
            # align the data by these integer shifts
            for t in range(len(dt)):
                ib = imax == t
                Fg[:, :, ib] = np.roll(Fg[:, :, ib], dt[t], axis=0)
                dall[iteration, ib] = dt[t]
            # new target frame based on our current best alignment
            F0 = np.mean(Fg, axis=2)[:, :, np.newaxis]

    # now we figure out how to split the probe into nblocks pieces
    # if nblocks = 1, then we're doing rigid registration
    nybins = F.shape[0]
    yl = np.floor(nybins / nblocks).astype("int") - 1
    # MATLAB rounds 0.5 to 1. Python uses "Bankers Rounding".
    # Numpy uses round to nearest even. Force the result to be like MATLAB
    # by adding a tiny constant.
    ifirst = np.round(np.linspace(0, nybins - yl - 1, 2 * nblocks - 1) + 1e-10).astype(
        "int"
    )
    ilast = ifirst + yl  # 287

    ##

    nblocks = len(ifirst)
    yblk = np.zeros(len(ifirst))

    # for each small block, we only look up and down this many samples to find
    # nonrigid shift
    n = 5
    dt = np.arange(-n, n + 1)

    # this part determines the up/down covariance for each block without
    # shifting anything
    dcs = np.zeros((2 * n + 1, Nbatches, nblocks))
    for j in range(nblocks):
        isub = np.arange(ifirst[j], ilast[j]+1)
        yblk[j] = np.mean(ysamp[isub])
        Fsub = Fg[isub, :, :]
        for t in range(len(dt)):
            Fs = np.roll(Fsub, dt[t], axis=0)
            dcs[t, :, j] = np.mean(Fs * F0[isub, :, :], axis=(0,1))

    # to find sub-integer shifts for each block ,
    # we now use upsampling, based on kriging interpolation
    dtup = np.linspace(-n, n, (2 * n * 10) + 1)
    K = kernelD(
        dt[np.newaxis, :], dtup[np.newaxis], 1
    )  # this kernel is fixed as a variance of 1
    dcs = cp.array(dcs)
    # dcs = my_conv2_cpu(dcs, .5, [0,1,2])
    for i in range(dcs.shape[0]):
        dcs[i, :, :] = my_conv2_cpu(
            dcs[i, :, :], 0.5, [0, 1]
        )  # some additional smoothing for robustness, across all dimensions
    for i in range(dcs.shape[2]):
        dcs[:, :, i] = my_conv2_cpu(
            dcs[:, :, i], 0.5, [0]
        )  # some additional smoothing for robustness, across all dimensions
        # dcs = my_conv2(cp.array(dcs), .5, [1, 2]) # some additional smoothing for robustness, across all dimensions
    dcs = dcs.get()

    # return K, dcs, dt, dtup
    imin = np.zeros((Nbatches, nblocks))
    for j in range(nblocks):
        # using the upsampling kernel K, get the upsampled cross-correlation
        # curves
        dcup = np.matmul(K.T, dcs[:, :, j])

        # find the max index of these curves
        imax = np.argmax(dcup, axis=0)

        # add the value of the shift to the last row of the matrix of shifts
        # (as if it was the last iteration of the main rigid loop )
        dall[niter - 1, :] = dtup[imax]

        # the sum of all the shifts equals the final shifts for this block
        imin[:, j] = np.sum(dall, axis=0)

    return imin, yblk, F0


def extended(ysamp, n, diff=None):
    if diff is None:
        diff = ysamp[1] - ysamp[0]
    pre = [ysamp[0] - i * diff for i in range(n, 0, -1)]
    post = [ysamp[-1] + i * diff for i in range(1, n)]
    return np.concatenate([pre, ysamp, post])


def zero_pad(shifts_in, n):
    pre = [0 for i in range(n, 0, -1)]
    post = [0 for i in range(1, n)]
    return np.concatenate([pre, shifts_in, post])


def kernel2D(xp, yp, sig):
    distx = np.abs(xp[:, 0] - yp[:, 0][np.newaxis, :].T).T
    disty = np.abs(xp[:, 1] - yp[:, 1][np.newaxis, :].T).T

    sigx = sig
    sigy = 1.5 * sig

    p = 1
    K = np.exp(-((distx / sigx) ** p) - (disty / sigy) ** p)

    return K


def shift_data(data, shift_matrix):
    """
    Applies the shift transformation to the data via matrix multiplication
    :param data: Data matrix to be shifted, numpy memmap array, (n_time, n_channels)
                dtype int16, f-contiguous
    :param shift_matrix: Tranformation matrix, numpy array, (n_channels, n_channels)
                dtype float64, c-contiguous
    :return: Shifted data, numpy array, (n_time, n_channels)
                dtype int16, f-contiguous
    """

    data_shifted = np.asfortranarray((data @ shift_matrix.T).astype("int16"))

    return data_shifted


def interpolate_1D(sample_shifts, sample_coords, probe_coords):
    """
    Interpolates the shifts found in one dimension to estimate the shifts for each channel
    :param sample_shifts: Detected shifts, numpy array
    :param sample_coords: Coordinates at which the detected shifts were found, numpy array
    :param probe_coords: Coordinates of the probe channels, numpy array
    :return: Upsampled shifts for each channel, numpy array
    """

    assert len(sample_coords) == len(sample_shifts)

    if len(sample_coords) == 1:
        return np.full(len(probe_coords), sample_shifts[0])

    else:
        # zero pad input so "extrapolation" tends to zero
        # MATLAB uses a "modified Akima" which is proprietry :(
        # _ysamp = extended(ysamp, 2, 10000)
        # _shifts_in = zero_pad(shifts_in, 2)
        # interpolation_function = Akima1DInterpolator(_ysamp, _shifts_in)
        interpolation_function = Akima1DInterpolator(sample_coords, sample_shifts)

        # interpolation_function = interp1d(ysamp, shifts_in, kind='cubic', fill_value=([0],[0])) #'extrapolate')
        shifts = interpolation_function(probe_coords, nu=0, extrapolate="True")

        return shifts


def get_kernel_matrix(probe, shifts, sig):
    """
    Calculate kernel prediction matrix for Gaussian Kriging interpolation
    :param probe: Bunch object with individual numpy arrays for the channel coordinates
    :param shifts: Numpy array of the estimated shift for each channel
    :param sig: Standard deviation used in Gaussian interpolation, float value
    :return: Prediction matrix, numpy array
    """

    # 2D coordinates of the original channel positions
    coords_old = np.vstack([probe.xc, probe.yc]).T

    # 2D coordinates of the new channel positions
    coords_new = np.copy(coords_old)
    coords_new[:, 1] = coords_new[:, 1] - shifts

    # 2D kernel of the original channel positions
    Kxx = kernel2D(coords_old, coords_old, sig)

    # 2D kernel of the new channel positions
    Kyx = kernel2D(coords_new, coords_old, sig)

    # kernel prediction matrix
    prediction_matrix = Kyx @ np.linalg.pinv(Kxx + 0.01 * np.eye(Kxx.shape[0]))

    return prediction_matrix


def apply_drift_transform(dat, shifts_in, ysamp, probe, sig):
    """
    Apply kriging interpolation on data batch
    :param dat: Data batch, (n_time, n_channels)
    :param shifts_in: Shifts per block (n_blocks)
    :param ysamp: Y-coords for block centres (n_blocks)
    :param probe: Bunch object with xc (xcoords) and yc (ycoords) attributes
    :param sig: Standard deviation for Gaussian in kriging interpolation
    :return: Shifted data batch via a kriging transformation
    """

    # upsample to get shifts for each channel
    shifts = interpolate_1D(shifts_in, ysamp, probe.yc)

    # kernel prediction matrix
    kernel_matrix = get_kernel_matrix(probe, shifts, sig)

    # apply shift transformation to the data
    data_shifted = shift_data(dat, kernel_matrix)

    return data_shifted


def shift_batch_on_disk2(
    ibatch,
    shifts_in,
    ysamp,
    sig,
    probe,
    data_loader,
):

    # load the batch
    dat = data_loader.load_batch(ibatch, rescale=False)

    # upsample the shift for each channel using interpolation
    shifts = interpolate_1D(shifts_in, ysamp, probe.yc)

    # kernel prediction matrix
    kernel_matrix = get_kernel_matrix(probe, shifts, sig)

    # apply shift transformation to the data
    data_shifted = shift_data(dat, kernel_matrix)

    # write the aligned data back to the same file
    data_loader.write_batch(ibatch, data_shifted)


def standalone_detector(wTEMP, wPCA, NrankPC, yup, xup, Nbatch, data_loader, probe, params):
    """
    Detects spikes across the entire recording using generic templates.
    Each generic template has rank one (in space-time).
    In time, we use the 1D template prototypes found in wTEMP.
    In space, we use Gaussian weights of several sizes, centered
    on (x,y) positions that are part of a super-resolution grid covering the
    entire probe (pre-specified in the calling function).
    In total, there ~100x more generic templates than channels.
    """

    # minimum/base sigma for the Gaussian.
    sig = 10

    # grid of centers for the generic tempates
    ycup, xcup = np.meshgrid(yup, xup)

    # Get nearest channels for every template center.
    # Template products will only be computed on these channels.
    NchanNear = 10
    iC, dist = getClosestChannels2(ycup, xcup, probe.yc, probe.xc, NchanNear)

    # Templates with centers that are far from an active site are discarded
    dNearActiveSite = np.median(np.diff(np.unique(probe.yc)))
    igood = dist[0, :] < dNearActiveSite
    iC = iC[:, igood]
    dist = dist[:, igood]
    ycup = cp.array(ycup).T.ravel()[igood]
    xcup = cp.array(xcup).T.ravel()[igood]

    # number of nearby templates to compare for local template maximum
    NchanNearUp = 10 * NchanNear
    iC2, dist2 = getClosestChannels2(ycup, xcup, ycup, xcup, NchanNearUp)

    # pregenerate the Gaussian weights used for spatial components
    nsizes = 5
    v2 = cp.zeros((5, dist.shape[1]), dtype=np.float32)
    for ibatch in range(0, nsizes):
        v2[ibatch, :] = np.sum(np.exp(-2 * (dist ** 2) / (sig * (ibatch + 1)) ** 2), 0)

    # build up Params
    NchanUp = iC.shape[1]
    Params = (
        params.NT,
        probe.Nchan,
        params.nt0,
        NchanNear,
        NrankPC,
        params.nt0min,
        params.genericSpkTh,
        NchanUp,
        NchanNearUp,
        sig,
    )

    # preallocate the results we assume 50 spikes per channel per second max
    rl = data_loader.data.shape[0] / params.fs / probe.Nchan   # record length
    st3 = np.zeros((int(np.ceil(rl * 50 * probe.Nchan)), 5))
    st3[:, 4] = -1  # batch_id can be zero
    t0 = 0
    nsp = 0  # counter for total number of spikes

    pbar = trange(Nbatch, desc="Detecting Spikes, 0 batches, 0 spikes", leave=True)
    for ibatch in pbar:
        # get a batch of whitened and filtered data
        dataRAW = data_loader.load_batch(ibatch)

        # run the CUDA function on this batch
        dat, kkmax, st, cF = spikedetector3(
            Params, dataRAW, wTEMP, iC, dist, v2, iC2, dist2
        )
        # upsample the y position using the center of mass of template products
        # coming out of the CUDA function.
        ys = probe.yc[cp.asnumpy(iC)]
        cF0 = np.maximum(cF, 0)
        cF0 = cF0 / np.sum(cF0, 0)
        iChan = st[1, :]
        yct = np.sum(cF0 * ys[:, iChan], 0)

        # build st for the current batch
        st[1, :] = yct

        # add time offset to get correct spike times
        toff = t0 + params.nt0min + params.NT * ibatch
        st[0, :] = st[0, :] + toff

        # st[4, :] = k  # add batch number
        st = np.concatenate([st, np.full((1, st.shape[1]), ibatch)])

        nsp0 = st.shape[1]
        if nsp0 + nsp > st3.shape[0]:
            # Pre-allocate more space if needed
            st3 = np.concatenate((st3, np.zeros((1000000, 5))), axis=0)

        st3[nsp : nsp0 + nsp, :] = st.T
        nsp = nsp + nsp0

        if ibatch % 10 == 0 or ibatch == (Nbatch - 1):
            pbar.set_description(f"Detecting Spikes, {ibatch+1} batches, {nsp} spikes", refresh=True)

    spikes = Bunch()
    spikes.times = st3[:nsp, 0]
    spikes.depths = st3[:nsp, 1]
    spikes.amps = st3[:nsp, 2]
    spikes.batches = st3[:nsp, 4]

    return spikes


def get_drift(spikes, probe, Nbatches, nblocks=5, genericSpkTh=10):
    """
    Estimates the drift using the spiking activity found in the first pass through the data
    :param spikes: Bunch object, contains the depths, amplitudes, times and batches of the spikes.
                    Each attribute is stored as a 1d numpy array
    :param probe: Bunch object, contains the x and y coordinates stored as 1D numpy arrays
    :param Nbatches: No batches in the dataset
    :param nblocks: No of blocks to divide the probe into when estimating drift
    :param genericSpkTh: Min amplitude of spiking activity found
    :return: dshift: 2D numpy array of drift estimates per batch and per sub-block in um
                    size (Nbatches, 2*nblocks-1)
            yblk: 1D numpy array containing average y position of each sub-block
    """

    # binning width across Y (um)
    dd = 5

    # min and max for the range of depths
    dmin = min(probe.yc) - 1

    dmax = int(1 + np.ceil((max(probe.yc) - dmin) / dd))

    # preallocate matrix of counts with 20 bins, spaced logarithmically
    F = np.zeros((dmax, 20, Nbatches))
    for t in range(Nbatches):
        # find spikes in this batch
        ix = np.where(spikes.batches == t)[0]

        # subtract offset
        spike_depths_batch = spikes.depths[ix] - dmin

        # amplitude bin relative to the minimum possible value
        spike_amps_batch = np.log10(np.clip(spikes.amps[ix], None, 99)) - np.log10(genericSpkTh)
        # normalization by maximum possible value
        spike_amps_batch = spike_amps_batch / (np.log10(100) - np.log10(genericSpkTh))

        # multiply by 20 to distribute a [0,1] variable into 20 bins
        # sparse is very useful here to do this binning quickly
        i, j, v, m, n = (
            np.ceil(1e-5 + spike_depths_batch / dd).astype("int"),
            np.minimum(np.ceil(1e-5 + spike_amps_batch * 20), 20).astype("int"),
            np.ones(len(ix)),
            dmax,
            20,
        )
        M = coo_matrix((v, (i-1, j-1)), shape=(m,n)).toarray()

        # the counts themselves are taken on a logarithmic scale (some neurons
        # fire too much!)
        F[:, :, t] = np.log2(1 + M)

    ysamp = dmin + dd * np.arange(1, dmax+1) - dd / 2
    imin, yblk, F0 = align_block2(F, ysamp, nblocks)

    # convert to um
    dshift = imin * dd

    return dshift, yblk


def datashift2(ctx):
    """
    Main function to re-register the preprocessed data
    """
    params = ctx.params
    probe = ctx.probe
    raw_data = ctx.raw_data
    ir = ctx.intermediate
    Nbatch = ir.Nbatch

    ir.xc, ir.yc = probe.xc, probe.yc

    # The min and max of the y and x ranges of the channels
    ymin = min(ir.yc)
    ymax = max(ir.yc)
    xmin = min(ir.xc)
    xmax = max(ir.xc)

    # Determine the average vertical spacing between channels.
    # Usually all the vertical spacings are the same, i.e. on Neuropixels probes.
    dmin = np.median(np.diff(np.unique(ir.yc)))
    logger.info(f"pitch is {dmin} um")
    yup = np.arange(
        start=ymin, step=dmin / 2, stop=ymax + (dmin / 2)
    )  # centers of the upsampled y positions

    # Determine the template spacings along the x dimension
    x_range = xmax - xmin
    npt = floor(
        x_range / 16
    )  # this would come out as 16um for Neuropixels probes, which aligns with the geometry.
    xup = np.linspace(xmin, xmax, npt + 1)  # centers of the upsampled x positions

    # Set seed
    if params.seed:
        np.random.seed(params.seed)

    # determine prototypical timecourses by clustering of simple threshold crossings.
    wTEMP, wPCA = extractTemplatesfromSnippets(
        data_loader=ir.data_loader, probe=probe, params=params, Nbatch=Nbatch
    )

    # Extract all the spikes across the recording that are captured by the
    # generic templates. Very few real spikes are missed in this way.
    spikes = standalone_detector(
        wTEMP, wPCA, params.nPCs, yup, xup, Nbatch, ir.data_loader, probe, params
    )

    if params.save_drift_spike_detections:
        drift_path = ctx.context_path / 'drift'
        if not os.path.isdir(drift_path):
            os.mkdir(drift_path)
        np.save(drift_path / 'spike_times.npy', spikes.times)
        np.save(drift_path / 'spike_depths.npy', spikes.depths)
        np.save(drift_path / 'spike_amps.npy', spikes.amps)

    dshift, yblk = get_drift(spikes, probe, Nbatch, params.nblocks, params.genericSpkTh)

    # sort in case we still want to do "tracking"
    iorig = np.argsort(np.mean(dshift, axis=1))

    # register the data batch by batch
    for ibatch in tqdm(range(Nbatch), desc='Shifting Data'):

        # load the batch from binary file
        dat = ir.data_loader.load_batch(ibatch, rescale=False)

        # align via kriging interpolation
        data_shifted = apply_drift_transform(dat, dshift[ibatch, :], yblk, probe, params.sig_datashift)

        # write the aligned data back to the same file
        ir.data_loader.write_batch(ibatch, data_shifted)

    logger.info(f"Shifted up/down {Nbatch} batches")

    return Bunch(iorig=iorig, dshift=dshift, yblk=yblk)



import logging
import os

if os.getenv('MOCK_CUPY', False):
    from pykilosort.testing.mock_cupy import cupy 
    from pykilosort.testing.mock_cupyx import cupyx
else:
    import cupy
    import cupyx

from .utils import Bunch, memmap_binary_file, read_data, load_probe, plot_dissimilarity_matrices, plot_diagnostics
from .main import run, run_export, run_spikesort, run_preprocess
from .io.probes import np1_probe, np2_probe, np2_4shank_probe


__version__ = '1.4.6'


# Set a null handler on the root logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


_logger_fmt = '%(asctime)s.%(msecs)03d [%(levelname)s] %(caller)s %(message)s'
_logger_date_fmt = '%H:%M:%S'


class _Formatter(logging.Formatter):
    def format(self, record):
        # Only keep the first character in the level name.
        record.levelname = record.levelname[0]
        filename = os.path.splitext(os.path.basename(record.pathname))[0]
        record.caller = '{:s}:{:d}'.format(filename, record.lineno).ljust(20)
        message = super(_Formatter, self).format(record)
        color_code = {'D': '90', 'I': '0', 'W': '33', 'E': '31'}.get(record.levelname, '7')
        message = '\33[%sm%s\33[0m' % (color_code, message)
        return message


def add_default_handler(level='INFO', logger=logger, filename=None):
    if filename is None:
        handler = logging.StreamHandler()
    else:
        handler = logging.FileHandler(filename)
    handler.setLevel(level)
    formatter = _Formatter(fmt=_logger_fmt, datefmt=_logger_date_fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)



import numpy as np

from pykilosort.datashift2 import shift_data, align_block2, apply_drift_transform
from pykilosort.utils import Bunch

def test_shift_data():
    data = np.asfortranarray(np.random.randint(-100, 100, size=(65000, 384), dtype=np.int16))
    transformation = np.random.normal(size=(384, 384))
    test_result = shift_data(data, transformation)
    assert type(test_result == np.ndarray)
    assert test_result.dtype == np.int16
    assert np.isfortran(test_result)
    assert np.allclose(test_result, np.asfortranarray((data @ transformation.T).astype("int16")))


def test_align_block2():
    # Rigid Case
    spike_histograms = np.zeros((20, 2, 3))
    spike_histograms[2:4,:,0] = 1
    spike_histograms[3:5,:,1] = 1
    spike_histograms[4:6,:,2] = 1
    imin, yblk, F0 = align_block2(spike_histograms, np.arange(20), 1)
    imin -= np.mean(imin)
    assert np.allclose(imin, np.array([[1],[0],[-1]]))


def test_apply_drift_transform():

    # Rigid Case
    dat = np.zeros((50,10), order='F')
    dat[:,4] = 10
    shifts_in = np.array([1])
    ysamp = np.array([5])
    probe = Bunch()
    probe.yc = np.arange(10)
    probe.xc = np.zeros(10)
    sig = 1

    data_shifted = apply_drift_transform(dat, shifts_in, ysamp, probe, sig)
    assert np.argmax(np.mean(data_shifted, axis=0)) == 5



import math

from pykilosort import cupy as cu
import numpy as np

from pykilosort import cluster, utils


def test_get_closest_channels():
    probe = utils.Bunch(
            {
                "NchanTOT": 4,
                "chanMap": np.array([0, 1, 2, 3]),
                "kcoords": np.array([1.0, 1.0, 1.0, 1.0]),
                "xc": np.array([0.0, 0.0, 1.0, 1.0]),
                "yc": np.array([0.0, 1.0, 0.0, 1.0]),
                "Nchan": 4,
                }
            )
    sigma = 30
    NchanClosest = 4

    iC, mask, C2C = cluster.getClosestChannels(probe, sigma, NchanClosest)

    assert cu.all(iC == cu.array([
        [0, 1, 2, 3],
        [1, 0, 0, 1],
        [2, 3, 3, 2],
        [3, 2, 1, 0]
        ]))

    assert cu.allclose([0.5], mask, atol=0.05)

    assert cu.allclose(math.sqrt(2), cu.fliplr(C2C).diagonal())


def test_sort_batches2():
    ccb0 = cu.random.rand(10,10)
    ccb1, isort = cluster.sortBatches2(ccb0)

    assert cu.all(ccb1 == ccb0[isort, :][:, isort])



from pathlib import Path
import numpy as np
from scipy.signal import lfilter as lfilter_cpu
import cupy as cp

from pykilosort.postprocess import my_conv2, merge_by_order

sig = 250
tmax = 1000

test_path = Path(__file__).parent


def test_convolve_real_data():
    file_s1 = test_path.joinpath('my_conv2_input.npy')
    file_expected = test_path.joinpath('my_conv2_output.npy')
    if not file_s1.exists() or not file_expected.exists():
        return

    s1 = np.load(file_s1)
    s1_expected = np.load(file_expected)

    def diff(out):
        return np.max(np.abs(cp.asnumpy(out[1000:-1000, :1]) - s1_expected[1000:-1000, :1]))

    x = cp.asarray(s1)
    assert diff(my_conv2(x, 250, 0, nwin=0, pad=None)) < 1e-6
    assert diff(my_conv2(x, 250, 0, nwin=0, pad='zeros')) < 1e-6
    assert diff(my_conv2(x, 250, 0, nwin=0, pad='constant')) < 1e-6

    assert diff(my_conv2(x, 250, 0, nwin=10000, pad=None)) < 1e-3
    assert diff(my_conv2(x, 250, 0, nwin=10000, pad='zeros')) < 1e-3
    assert diff(my_conv2(x, 250, 0, nwin=10000, pad='constant')) < 1e-3

    assert diff(my_conv2(x, 250, 0, nwin=0, pad='flip')) < 1e-6
    assert diff(my_conv2(x, 250, 0, nwin=10000, pad='flip')) < 1e-3


def create_test_dataset():
    cp = np  # cpu mode only here
    s1 = np.load(test_path.joinpath('my_conv2_input.npy'))
    s0 = np.copy(s1)
    tmax = np.ceil(4 * sig)
    dt = cp.arange(-tmax, tmax + 1)
    gauss = cp.exp(-dt ** 2 / (2 * sig ** 2))
    gauss = (gauss / cp.sum(gauss)).astype(np.float32)

    cNorm = lfilter_cpu(gauss, 1., np.r_[np.ones(s1.shape[0]), np.zeros(int(tmax))])
    cNorm = cNorm[int(tmax):]

    s1 = lfilter_cpu(gauss, 1, np.r_[s1, np.zeros((int(tmax), s1.shape[1]))], axis=0)
    s1 = s1[int(tmax):] / cNorm[:, np.newaxis]

    np.save(test_path.joinpath('my_conv2_output.npy'), s1)


def test_merge_by_order():

    n_chan = 32
    n_rank = 3
    n1 = 20
    n2 = 30

    total = n1 + n2

    features1 = np.asfortranarray(np.random.randn(n_chan, n_rank, n1))
    features2 = np.asfortranarray(np.random.randn(n_chan, n_rank, n2))

    times1 = np.sort(np.random.choice(total, n1, replace=False))
    mask = np.ones(total, dtype='bool')
    mask[times1] = False
    times2 = np.where(mask)[0]

    output = merge_by_order(features1, features2, times1, times2, axis=2)

    test_output = np.zeros((n_chan, n_rank, total), dtype=features1.dtype, order='F')
    for i in range(total):
        if i in times1:
            location = np.where(times1 == i)[0][0]
            test_output[:,:,i] = features1[:,:,location]
        else:
            location = np.where(times2 == i)[0][0]
            test_output[:,:,i] = features2[:,:,location]

    assert np.allclose(output, test_output)



import spikeglx
from pykilosort.preprocess import get_data_covariance_matrix
from pykilosort import ibl, run
from pykilosort import params
from neurodsp import voltage
params = ibl.ibl_pykilosort_params(1)
probe = params['probe']

from one.api import ONE

one = ONE(base_url="https://alyx.internationalbrainlab.org", cache_dir="/mnt/s1/pykilosort_reruns")
pid = 'f3988aec-b8c7-47d9-b1a0-5334eed7f92c'
pid = '3c283107-7012-48fc-a6c2-ed096b23974f'


eid, pname = one.pid2eid(pid)
session_path = one.eid2path(eid)
cbin_file = next(session_path.joinpath('raw_ephys_data', pname).glob('*.ap.cbin'))

sr = spikeglx.Reader(cbin_file)
butter_kwargs = {'N': 3, 'Wn': 300 / sr.fs * 2, 'btype': 'highpass'}


import scipy
from viewephys.gui import viewephys
t0 = 1254
first, last = (int(sr.fs * t0), int(sr.fs * (t0 + 1)))
butter_kwargs = {'N': 3, 'Wn': 300 / sr.fs * 2, 'btype': 'highpass'}
sos = scipy.signal.butter(**butter_kwargs, output='sos')

raw = sr[first:last, :-sr.nsync].T
butt = scipy.signal.sosfiltfilt(sos, raw)
destripe = voltage.destripe(raw, sr.fs, neuropixel_version=1, channel_labels=True)


h = sr.geometry
viewephys(butt, sr.fs, channels=sr.geometry, title='butter')
viewephys(destripe, sr.fs, channels=sr.geometry, title='destripe')
## %%

butter_kwargs = {'N': 3, 'Wn': 0.4, 'btype': 'highpass'}

kfilt = voltage.kfilt(destripe, collection=h['col'], ntr_pad=0, ntr_tap=None, lagc=None, butter_kwargs=butter_kwargs)
viewephys(kfilt, sr.fs, channels=h, title='kfilt')
viewephys(destripe - kfilt, sr.fs, channels=h, title='diff')

## %%


import numpy as np

def spiking_decon(x, nsop=36, white=.1):
    """
    Wiener deconvolution using predictive filters
    :param x:
    :param white:
    :return:
    """
    nc, ns = x.shape
    dec = x.copy()
    op = np.zeros((nc, nsop))
    for i in np.arange(nc):
        ac = np.fft.irfft(np.fft.rfft(x[i, :]) * np.conj(np.fft.rfft(x[i, :])))
        # ac = ac / ac[0]
        gtg = scipy.linalg.toeplitz(ac[:nsop])
        gtg = gtg + np.eye(nsop) * white * np.trace(gtg) / nsop
        gty = np.zeros(nsop)
        gty[0] = 1
        op = np.linalg.solve(gtg, gty)
        dec[i, :] = np.convolve(x[i, :], op, mode='full')[:ns]
    return dec

decon = spiking_decon(destripe.T, nsop=36, white=.1).T

viewephys(decon, sr.fs, channels=h, title='decon')

#
# import matplotlib.pyplot as plt
#
# plt.figure()
# plt.plot(x)
# plt.plot(dec)
# plt.show()
# plt.figure()
#
# plt.psd(x)
# plt.psd(dec)
# plt.show()



import numpy as np
from numpy.testing import assert_allclose as ac
from scipy.signal import lfilter as lfilter_cpu
import cupy as cp

import pytest

from pykilosort.cptools import (
    median, lfilter, svdecon, svdecon_cpu, free_gpu_memory,
    convolve_cpu, convolve_gpu, convolve_gpu_direct, convolve_gpu_chunked)


def test_median_1(dtype, axis):
    arr = cp.random.rand(2000, 1000).astype(dtype)
    m1 = cp.asnumpy(median(arr, axis=axis))
    m2 = np.median(cp.asnumpy(arr), axis=axis)
    assert np.allclose(m1, m2)


@pytest.mark.requires_gpu
def test_lfilter_1():
    tmax = 1000
    dt = np.arange(-tmax, tmax + 1)
    gaus = np.exp(-dt ** 2 / (2 * 250 ** 2))
    b = gaus / np.sum(gaus)

    a = 1.

    n = 2000
    arr = np.r_[np.ones(n), np.zeros(n)]

    fil_gpu = cp.asnumpy(lfilter(b, a, cp.asarray(arr), axis=0)).ravel()
    fil_cpu = lfilter_cpu(b, a, arr, axis=0)

    assert np.allclose(fil_cpu, fil_gpu, atol=1e-6)


@pytest.mark.requires_gpu
def test_lfilter_2():
    b = (0.96907117, -2.90721352, 2.90721352, -0.96907117)
    a = (1., -2.93717073, 2.87629972, -0.93909894)
    arr = np.random.rand(1000, 100).astype(np.float32)

    fil_gpu = cp.asnumpy(lfilter(b, a, cp.asarray(arr), axis=0))
    fil_cpu = lfilter_cpu(b, a, arr, axis=0)

    assert np.allclose(fil_cpu, fil_gpu, atol=.2)


def test_svdecon_1():
    X = cp.random.rand(10, 10)

    U, S, V = svdecon(X)
    Un, Sn, Vn = svdecon_cpu(X)

    assert np.allclose(S, Sn)


@pytest.fixture(params=(100, 2_000, 10_000, 50_000, 250_000))
def arr(request):
    return np.random.randn(int(request.param), 100)


@pytest.fixture
def gaus():
    dt = np.arange(-1000, 1000 + 1)
    gaus = np.exp(-dt ** 2 / (2 * 250 ** 2))
    return gaus / np.sum(gaus)


@pytest.fixture(params=(None, 'zeros', 'flip', 'constant'))
def pad(request):
    return request.param


@pytest.fixture
def nwin():
    return 2500


@pytest.mark.requires_gpu
def test_convolve(arr, gaus, pad, nwin):
    free_gpu_memory()
    npad = gaus.shape[0] // 2

    # Upload the arrays to the GPU.
    arr_gpu = cp.asarray(arr)
    gaus_gpu = cp.asarray(gaus)

    # Compute the convolution on the CPU.
    conv_cpu = convolve_cpu(arr, gaus)

    def check(y):
        ac(cp.asnumpy(y)[npad:-npad, :], conv_cpu[npad:-npad, :], atol=1e-3)

    # Check the GPU direct version with the CPU version.
    check(convolve_gpu_direct(arr_gpu, gaus_gpu, pad=pad))

    if arr.shape[0] > nwin:
        # Check the GPU chunked version with the CPU version.
        y = convolve_gpu_chunked(arr_gpu, gaus_gpu, pad=pad, nwin=nwin)
        check(y)

        # DEBUG
        # import matplotlib.pyplot as plt
        # plt.plot(cp.asnumpy(conv_cpu[npad:-npad, 0]))
        # plt.plot(cp.asnumpy(y[npad:-npad, 0]))
        # plt.show()

    if pad is None:
        # This function should automatically route to the correct GPU implementation.
        check(convolve_gpu(arr_gpu, gaus_gpu))



import numpy as np
from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_allclose as ac
from pytest import raises

from pykilosort.utils import Context, LargeArrayWriter, memmap_large_array, save_large_array, NpyWriter


def test_context_1(tmp_path):
    arr = np.random.randn(10, 20)

    c = Context(tmp_path)
    c.intermediate.test1 = arr

    ae(c.read('test1'), arr)

    c.write(test1=arr)
    ae(c.read('test1'), arr)

    c.write(test2=12.34)
    assert c.read('test2') == 12.34


def test_context_2(tmp_path):
    arr = np.random.randn(10, 20)

    c = Context(tmp_path)

    c.intermediate.test1 = arr
    c.intermediate.test2 = 12.34
    c.save()

    c.load()
    ae(c.intermediate.test1, arr)
    assert c.intermediate.test2 == 12.34

    c = Context(tmp_path)
    assert 'test1' not in c.intermediate
    c.load()
    ae(c.intermediate.test1, arr)
    assert c.intermediate.test2 == 12.34


def test_large_array_1(tmp_path):
    path = tmp_path / 'arr.dat'
    arr = np.random.rand(1000, 13).T

    law = LargeArrayWriter(path, dtype=arr.dtype, shape=(13, -1))
    with raises(AssertionError):
        law.append(np.random.rand(1000, 13))
    law.append(arr)
    law.close()

    arr_1 = memmap_large_array(path)
    assert arr_1.shape == arr.shape
    ac(arr_1, arr)


def test_large_array_2(tmp_path):
    path = tmp_path / 'arr.dat'
    arr = np.random.rand(1000, 13).T

    law = LargeArrayWriter(path, dtype=arr.dtype, shape=(13, -1))
    k = 10
    for i in range(arr.shape[-1] // k):
        law.append(arr[:, k * i: k * (i + 1)])
    law.close()

    arr_1 = memmap_large_array(path)
    assert arr_1.shape == arr.shape
    ac(arr_1, arr)


def test_large_array_3(tmp_path):
    path = tmp_path / 'arr.dat'

    law = LargeArrayWriter(path, dtype=np.float32, shape=(13, -1))
    arrs = []
    for i in range(100):
        arr = np.random.rand(np.random.randint(low=10, high=20), 13).T
        arrs.append(arr)
        law.append(arr)
    law.close()

    arr = np.concatenate(arrs, axis=1)
    arr_1 = memmap_large_array(path)
    assert arr_1.shape == arr.shape
    ac(arr_1, arr)


def test_save_large_array_1(tmp_path):
    path = tmp_path / 'arr.npy'
    for n in (0, 51, 99, 100, 101, 99, 1000, 1010):
        array = np.random.rand(n, 20, 30)
        with open(path, 'wb') as fp:
            save_large_array(fp, array)
        ac(np.load(path), array)


def test_npy_writer_1(tmp_path):
    path = tmp_path / 'arr.npy'
    for n in (0, 51, 99, 100, 101, 99, 1000, 1010):
        array = np.random.rand(n, 20, 30)
        nw = NpyWriter(path, array.shape, array.dtype)
        for row in array:
            nw.append(row)
        nw.close()
        ac(np.load(path), array)



import logging
from math import ceil

import numpy as np
import cupy as cp
from tqdm.auto import tqdm

from .preprocess import my_min, my_sum
from .cptools import svdecon, zscore, ones
from .utils import Bunch, get_cuda, plot_dissimilarity_matrices

logger = logging.getLogger(__name__)


def getClosestChannels(probe, sigma, NchanClosest):
    # this function outputs the closest channels to each channel,
    # as well as a Gaussian-decaying mask as a function of pairwise distances
    # sigma is the standard deviation of this Gaussian-mask

    # compute distances between all pairs of channels
    xc = cp.asarray(probe.xc, dtype=np.float32, order='F')
    yc = cp.asarray(probe.yc, dtype=np.float32, order='F')
    C2C = (xc[:, np.newaxis] - xc) ** 2 + (yc[:, np.newaxis] - yc) ** 2
    C2C = cp.sqrt(C2C)
    Nchan = C2C.shape[0]

    # sort distances
    isort = cp.argsort(C2C, axis=0)

    # take NchanCLosest neighbors for each primary channel
    iC = isort[:NchanClosest, :]

    # in some cases we want a mask that decays as a function of distance between pairs of channels
    # this is an awkward indexing to get the corresponding distances
    ix = iC + cp.arange(0, Nchan ** 2, Nchan)
    mask = cp.exp(-C2C.T.ravel()[ix] ** 2 / (2 * sigma ** 2))

    # masks should be unit norm for each channel
    mask = mask / cp.sqrt(1e-3 + cp.sum(mask ** 2, axis=0))

    return iC, mask, C2C


def isolated_peaks_new(S1, params):
    """
    takes a matrix of timepoints by channels S1
    outputs threshold crossings that are relatively isolated from other peaks
    outputs row, column and magnitude of the threshold crossing
    """
    S1 = cp.asarray(S1)

    # finding the local minimum in a sliding window within plus/minus loc_range extent
    # across time and across channels
    smin = my_min(S1, params.loc_range, [0, 1])

    # the peaks are samples that achieve this local minimum, AND have negativities less
    # than a preset threshold
    peaks = (S1 < smin + 1e-3) & (S1 < params.spkTh)

    # only take local peaks that are isolated from other local peaks
    # if there is another local peak close by, this sum will be at least 2
    sum_peaks = my_sum(peaks, params.long_range, [0, 1])
    # set to 0 peaks that are not isolated, and multiply with the voltage values
    peaks = peaks * (sum_peaks < 1.2) * S1

    # exclude temporal buffers
    peaks[:params.nt0, :] = 0
    peaks[-params.nt0:, :] = 0

    # find the non-zero peaks, and take their amplitudes
    col, row = cp.nonzero(peaks.T)
    # invert the sign of the amplitudes
    mu = -peaks[row, col]

    return row, col, mu


def get_SpikeSample(dataRAW, row, col, params):
    """
    given a batch of data (time by channels), and some time (row) and channel (col) indices for
    spikes, this function returns the 1D time clips of voltage around those spike times
    """
    nT, nChan = dataRAW.shape

    # times around the peak to consider
    dt = cp.arange(params.nt0)

    # the negativity is expected at nt0min, so we align the detected peaks there
    dt = -params.nt0min + dt

    # temporal indices (awkward way to index into full matrix of data)
    indsT = row + dt[:, np.newaxis] + 1  # broadcasting
    indsC = col

    indsC[indsC < 0] = 0  # anything that's out of bounds just gets set to the limit
    indsC[indsC >= nChan] = nChan - 1  # only needed for channels not time (due to time buffer)

    indsT = cp.transpose(cp.atleast_3d(indsT), [0, 2, 1])
    indsC = cp.transpose(cp.atleast_3d(indsC), [2, 0, 1])

    # believe it or not, these indices grab just the right timesamples forour spikes
    ix = indsT + indsC * nT

    # grab the data and reshape it appropriately (time samples  by channels by num spikes)
    clips = dataRAW.T.ravel()[ix[:, 0, :]].reshape((dt.size, row.size), order='F')  # HERE
    return clips


def extractPCfromSnippets(proc, probe=None, params=None, Nbatch=None):
    # extracts principal components for 1D snippets of spikes from all channels
    # loads a subset of batches to find these snippets

    NT = params.NT
    nPCs = params.nPCs
    Nchan = probe.Nchan

    batchstart = np.arange(0, NT * Nbatch + 1, NT).astype(np.int64)

    # extract the PCA projections
    # initialize the covariance of single-channel spike waveforms
    CC = cp.zeros(params.nt0, dtype=np.float32)

    # from every 100th batch
    for ibatch in range(0, Nbatch, 100):
        offset = Nchan * batchstart[ibatch]
        dat = proc.flat[offset:offset + NT * Nchan].reshape((-1, Nchan), order='F')
        if dat.shape[0] == 0:
            continue

        # move data to GPU and scale it back to unit variance
        dataRAW = cp.asarray(dat, dtype=np.float32) / params.scaleproc

        # find isolated spikes from each batch
        row, col, mu = isolated_peaks_new(dataRAW, params)

        # for each peak, get the voltage snippet from that channel
        c = get_SpikeSample(dataRAW, row, col, params)

        # scale covariance down by 1,000 to maintain a good dynamic range
        CC = CC + cp.dot(c, c.T) / 1e3

    # the singular vectors of the covariance matrix are the PCs of the waveforms
    U, Sv, V = svdecon(CC)

    wPCA = U[:, :nPCs]  # take as many as needed

    # adjust the arbitrary sign of the first PC so its negativity is downward
    # TODO: unclear - is 20 here the index into the spike waveform? Should this be hardcoded?
    #               - should it be nt0min instead?
    wPCA[:, 0] = -wPCA[:, 0] * cp.sign(wPCA[20, 0])

    return wPCA


def sortBatches2(ccb0):
    # takes as input a matrix of nBatches by nBatches containing
    # dissimilarities.
    # outputs a matrix of sorted batches, and the sorting order, such that
    # ccb1 = ccb0(isort, isort)

    # put this matrix on the GPU
    ccb0 = cp.asarray(ccb0, order='F')

    # compute its svd on the GPU (this might also be fast enough on CPU)
    u, s, v = svdecon(ccb0)
    # HACK: consistency with MATLAB
    u = u * cp.sign(u[0, 0])
    v = v * cp.sign(u[0, 0])

    # initialize the positions xs of the batch embeddings to be very small but proportional to
    # the first PC
    xs = .01 * u[:, 0] / cp.std(u[:, 0], ddof=1)

    # 200 iterations of gradient descent should be enough
    # TODO: move_to_config
    niB = 200

    # this learning rate should usually work fine, since it scales with the average gradient
    # and ccb0 is z-scored
    # TODO: move_to_config
    eta = 1
    for k in tqdm(range(niB), desc="Sorting %d batches" % ccb0.shape[0]):
        # euclidian distances between 1D embedding positions
        ds = (xs - xs[:, np.newaxis]) ** 2
        # the transformed distances go through this function
        W = cp.log(1 + ds)

        # the error is the difference between ccb0 and W
        err = ccb0 - W

        # ignore the mean value of ccb0
        err = err - cp.mean(err, axis=0)

        # backpropagate the gradients
        err = err / (1 + ds)
        err2 = err * (xs[:, np.newaxis] - xs)
        D = cp.mean(err2, axis=1)  # one half of the gradients is along this direction
        E = cp.mean(err2, axis=0)  # the other half is along this direction
        # we don't need to worry about the gradients for the diagonal because those are 0

        # final gradients for the embedding variable
        dx = -D + E.T

        # take a gradient step
        xs = xs - eta * dx

    # sort the embedding positions xs
    isort = cp.argsort(xs, axis=0)

    # sort the matrix of dissimilarities
    ccb1 = ccb0[isort, :][:, isort]
    
    return ccb1, isort


def initializeWdata2(call, uprojDAT, Nchan, nPCs, Nfilt, iC):
    # this function initializes cluster means for the fast kmeans per batch
    # call are time indices for the spikes
    # uprojDAT are features projections (Nfeatures by Nspikes)
    # some more parameters need to be passed in from the main workspace

    # pick random spikes from the sample
    # WARNING: replace ceil by floor because this is a random index, and 0/1 indexing
    # discrepancy between Python and MATLAB.
    irand = np.floor(np.random.rand(Nfilt) * uprojDAT.shape[1]).astype(np.int32)

    W = cp.zeros((nPCs, Nchan, Nfilt), dtype=np.float32)

    for t in range(Nfilt):
        ich = iC[:, call[irand[t]]]  # the channels on which this spike lives
        # for each selected spike, get its features
        W[:, ich, t] = uprojDAT[:, irand[t]].reshape(W[:, ich, t].shape, order='F')

    W = W.reshape((-1, Nfilt), order='F')  # HERE
    # add small amount of noise in case we accidentally picked the same spike twice
    W = W + .001 * cp.random.normal(size=W.shape).astype(np.float32)
    mu = cp.sqrt(cp.sum(W ** 2, axis=0))  # get the mean of the template
    W = W / (1e-5 + mu)  # and normalize the template
    W = W.reshape((nPCs, Nchan, Nfilt), order='F')  # HERE
    nW = (W[0, ...] ** 2)  # squared amplitude of the first PC feture
    W = W.reshape((nPCs * Nchan, Nfilt), order='F')  # HERE
    # determine biggest channel according to the amplitude of the first PC
    Wheights = cp.argmax(nW, axis=0)

    return W, mu, Wheights, irand


def mexThSpkPC(Params, dataRAW, wPCA, iC):
    code, constants = get_cuda('mexThSpkPC')
    Nthreads = constants.Nthreads
    maxFR = constants.maxFR

    NT, Nchan, NchanNear, nt0, nt0min, spkTh, NrankPC = Params
    NT = int(NT)
    Nchan = int(Nchan)

    # Input GPU arrays.
    d_Params = cp.asarray(Params, dtype=np.float64, order='F')
    d_data = cp.asarray(dataRAW, dtype=np.float32, order='F')
    d_W = cp.asarray(wPCA, dtype=np.float32, order='F')
    d_iC = cp.asarray(iC, dtype=np.int32, order='F')

    # New GPU arrays.
    d_dout = cp.zeros((Nchan, NT), dtype=np.float32, order='F')
    d_dmax = cp.zeros((Nchan, NT), dtype=np.float32, order='F')
    d_st = cp.zeros(maxFR, dtype=np.int32, order='F')
    d_id = cp.zeros(maxFR, dtype=np.int32, order='F')
    d_counter = cp.zeros(1, dtype=np.int32, order='F')

    # filter the data with the temporal templates
    Conv1D = cp.RawKernel(code, 'Conv1D')
    Conv1D((Nchan,), (Nthreads,), (d_Params, d_data, d_W, d_dout))

    # get the max of the data
    max1D = cp.RawKernel(code, 'max1D')
    max1D((Nchan,), (Nthreads,), (d_Params, d_dout, d_dmax))

    # take max across nearby channels
    maxChannels = cp.RawKernel(code, 'maxChannels')
    maxChannels(
        (int(NT // Nthreads),), (Nthreads,),
        (d_Params, d_dout, d_dmax, d_iC, d_st, d_id, d_counter))

    # move d_x to the CPU
    minSize = 1
    minSize = min(maxFR, int(d_counter[0]))

    d_featPC = cp.zeros((NrankPC * NchanNear, minSize), dtype=np.float32, order='F')

    d_id2 = cp.zeros(minSize, dtype=np.int32, order='F')

    if (minSize > 0):
        computeProjections = cp.RawKernel(code, 'computeProjections')
        computeProjections(
            (minSize,), (NchanNear, NrankPC), (d_Params, d_data, d_iC, d_st, d_id, d_W, d_featPC))

    # TODO: check that the copy occurs on the GPU only
    d_id2[:] = d_id[:minSize]

    # Free memory.
    # TODO: unclear - does this do something special for cupy objects?
    #               - all of these go out of scope in the next line anyway
    del d_st, d_id, d_counter, d_Params, d_dmax, d_dout
    # free_gpu_memory()

    return d_featPC, d_id2


def extractPCbatch2(proc, params, probe, wPCA, ibatch, iC, Nbatch):
    # this function finds threshold crossings in the data using
    # projections onto the pre-determined principal components
    # wPCA is number of time samples by number of PCs
    # ibatch is a scalar indicating which batch to analyze
    # iC is NchanNear by Nchan, indicating for each channel the nearest
    # channels to it

    nt0min = params.nt0min
    spkTh = params.ThPre
    nt0, NrankPC = wPCA.shape
    NT, Nchan = params.NT, probe.Nchan

    # starts with predefined PCA waveforms
    wPCA = wPCA[:, :3]

    NchanNear = iC.shape[0]

    # batches start at these timepoints
    batchstart = np.arange(0, NT * Nbatch + 1, NT).astype(np.int64)

    offset = Nchan * batchstart[ibatch]
    dat = proc.flat[offset:offset + NT * Nchan].reshape((-1, Nchan), order='F')
    dataRAW = cp.asarray(dat, dtype=np.float32) / params.scaleproc

    # another Params variable to take all our parameters into the C++ code
    Params = [NT, Nchan, NchanNear, nt0, nt0min, spkTh, NrankPC]

    # call a CUDA function to do the hard work
    # returns a matrix of features uS, as well as the center channels for each spike
    uS, idchan = mexThSpkPC(Params, dataRAW, wPCA, iC)

    return uS, idchan


def mexClustering2(Params, uproj, W, mu, call, iMatch, iC):

    code, _ = get_cuda('mexClustering2')

    Nspikes = int(Params[0])
    NrankPC = int(Params[1])
    Nfilters = int(Params[2])
    NchanNear = int(Params[6])
    Nchan = int(Params[7])

    d_Params = cp.asarray(Params, dtype=np.float64, order='F')
    d_uproj = cp.asarray(uproj, dtype=np.float32, order='F')
    d_W = cp.asarray(W, dtype=np.float32, order='F')
    d_mu = cp.asarray(mu, dtype=np.float32, order='F')
    d_call = cp.asarray(call, dtype=np.int32, order='F')
    d_iC = cp.asarray(iC, dtype=np.int32, order='F')
    d_iMatch = cp.asarray(iMatch, dtype=bool, order='F')

    d_dWU = cp.zeros((NrankPC * Nchan, Nfilters), dtype=np.float32, order='F')
    d_cmax = cp.zeros((Nspikes, Nfilters), dtype=np.float32, order='F')
    d_id = cp.zeros(Nspikes, dtype=np.int32, order='F')
    d_x = cp.zeros(Nspikes, dtype=np.float32, order='F')
    d_nsp = cp.zeros(Nfilters, dtype=np.int32, order='F')
    d_V = cp.zeros(Nfilters, dtype=np.float32, order='F')

    # get list of cmaxes for each combination of neuron and filter
    computeCost = cp.RawKernel(code, 'computeCost')
    computeCost(
        (Nfilters,), (1024,), (d_Params, d_uproj, d_mu, d_W, d_iMatch, d_iC, d_call, d_cmax))

    # loop through cmax to find best template
    bestFilter = cp.RawKernel(code, 'bestFilter')
    bestFilter((40,), (256,), (d_Params, d_iMatch, d_iC, d_call, d_cmax, d_id, d_x))

    # average all spikes for same template -- ORIGINAL
    average_snips_v2 = cp.RawKernel(code, 'average_snips_v2')
    average_snips_v2(
        (Nfilters,), (NrankPC, NchanNear), (d_Params, d_iC, d_call, d_id, d_uproj, d_cmax, d_dWU))

    count_spikes = cp.RawKernel(code, 'count_spikes')
    count_spikes((7,), (256,), (d_Params, d_id, d_nsp, d_x, d_V))

    del d_Params, d_V

    return d_dWU, d_id, d_x, d_nsp, d_cmax


def mexDistances2(Params, Ws, W, iMatch, iC, Wh, mus, mu):
    code, _ = get_cuda('mexDistances2')

    Nspikes = int(Params[0])
    Nfilters = int(Params[2])

    d_Params = cp.asarray(Params, dtype=np.float64, order='F')

    d_Ws = cp.asarray(Ws, dtype=np.float32, order='F')
    d_W = cp.asarray(W, dtype=np.float32, order='F')
    d_iMatch = cp.asarray(iMatch, dtype=bool, order='F')
    d_iC = cp.asarray(iC, dtype=np.int32, order='F')
    d_Wh = cp.asarray(Wh, dtype=np.int32, order='F')
    d_mu = cp.asarray(mu, dtype=np.float32, order='F')
    d_mus = cp.asarray(mus, dtype=np.float32, order='F')

    d_cmax = cp.zeros(Nspikes * Nfilters, dtype=np.float32, order='F')
    d_id = cp.zeros(Nspikes, dtype=np.int32, order='F')
    d_x = cp.zeros(Nspikes, dtype=np.float32, order='F')

    # get list of cmaxes for each combination of neuron and filter
    computeCost = cp.RawKernel(code, 'computeCost')
    computeCost(
        (Nfilters,), (1024,), (d_Params, d_Ws, d_mus, d_W, d_mu, d_iMatch, d_iC, d_Wh, d_cmax))

    # loop through cmax to find best template
    bestFilter = cp.RawKernel(code, 'bestFilter')
    bestFilter((40,), (256,), (d_Params, d_iMatch, d_Wh, d_cmax, d_mus, d_id, d_x))

    del d_Params, d_cmax

    return d_id, d_x


def clusterSingleBatches(ctx, sanity_plots=False, plot_widgets=None, plot_pos=0):
    """
    outputs an ordering of the batches according to drift
    for each batch, it extracts spikes as threshold crossings and clusters them with kmeans
    the resulting cluster means are then compared for all pairs of batches, and a dissimilarity
    score is assigned to each pair
    the matrix of similarity scores is then re-ordered so that low dissimilaity is along
    the diagonal
    """
    Nbatch = ctx.intermediate.Nbatch
    params = ctx.params
    probe = ctx.probe
    raw_data = ctx.raw_data
    ir = ctx.intermediate
    proc = ir.proc

    if not params.reorder:
        # if reordering is turned off, return consecutive order
        iorig = np.arange(Nbatch)
        return iorig, None, None

    nPCs = params.nPCs
    Nfilt = ceil(probe.Nchan / 2)

    # extract PCA waveforms pooled over channels
    wPCA = extractPCfromSnippets(proc, probe=probe, params=params, Nbatch=Nbatch)

    Nchan = probe.Nchan
    # TODO: move_to_config
    niter = 10  # iterations for k-means. we won't run it to convergence to save time

    nBatches = Nbatch
    NchanNear = min(Nchan, 2 * 8 + 1)

    # initialize big arrays on the GPU to hold the results from each batch
    # this holds the unit norm templates
    Ws = cp.zeros((nPCs, NchanNear, Nfilt, nBatches), dtype=np.float32, order='F')
    # this holds the scalings
    mus = cp.zeros((Nfilt, nBatches), dtype=np.float32, order='F')
    # this holds the number of spikes for that cluster
    ns = cp.zeros((Nfilt, nBatches), dtype=np.float32, order='F')
    # this holds the center channel for each template
    Whs = ones((Nfilt, nBatches), dtype=np.int32, order='F')

    i0 = 0
    # TODO: move_to_config
    NrankPC = 3  # I am not sure if this gets used, but it goes into the function

    # return an array of closest channels for each channel
    iC = getClosestChannels(probe, params.sigmaMask, NchanNear)[0]

    for ibatch in tqdm(range(nBatches), desc="Clustering spikes"):
        enough_spikes = False

        # extract spikes using PCA waveforms
        uproj, call = extractPCbatch2(
            proc, params, probe, wPCA, min(nBatches - 2, ibatch), iC, Nbatch)

        if cp.sum(cp.isnan(uproj)) > 0:
            break  # I am not sure what case this safeguards against....

        if uproj.shape[1] > Nfilt:
            enough_spikes = True

            # this initialize the k-means
            W, mu, Wheights, irand = initializeWdata2(call, uproj, Nchan, nPCs, Nfilt, iC)

            # Params is a whole bunch of parameters sent to the C++ scripts inside a float64 vector
            Params = [uproj.shape[1], NrankPC, Nfilt, 0, W.shape[0], 0, NchanNear, Nchan]

            for i in range(niter):

                Wheights = Wheights.reshape((1, 1, -1), order='F')
                iC = cp.atleast_3d(iC)

                # we only compute distances to clusters on the same channels
                # this tells us which spikes and which clusters might match
                iMatch = cp.min(cp.abs(iC - Wheights), axis=0) < .1

                # get iclust and update W
                # CUDA script to efficiently compute distances for pairs in which iMatch is 1
                dWU, iclust, dx, nsp, dV = mexClustering2(Params, uproj, W, mu, call, iMatch, iC)

                dWU = dWU / (1e-5 + nsp.T)  # divide the cumulative waveform by the number of spike

                mu = cp.sqrt(cp.sum(dWU ** 2, axis=0))  # norm of cluster template
                W = dWU / (1e-5 + mu)  # unit normalize templates

                W = W.reshape((nPCs, Nchan, Nfilt), order='F')
                nW = W[0, ...] ** 2  # compute best channel from the square of the first PC feature
                W = W.reshape((Nchan * nPCs, Nfilt), order='F')

                Wheights = cp.argmax(nW, axis=0)  # the new best channel of each cluster template

            # carefully keep track of cluster templates in dense format
            W = W.reshape((nPCs, Nchan, Nfilt), order='F')
            W0 = cp.zeros((nPCs, NchanNear, Nfilt), dtype=np.float32, order='F')
            for t in range(Nfilt):
                W0[..., t] = W[:, iC[:, Wheights[t]], t].squeeze()
            # I don't really know why this needs another normalization
            W0 = W0 / (1e-5 + cp.sum(cp.sum(W0 ** 2, axis=0)[np.newaxis, ...], axis=1) ** .5)

        # if a batch doesn't have enough spikes, it gets the cluster templates of the previous batc
        if enough_spikes:
            Ws[..., ibatch] = W0
            mus[:, ibatch] = mu
            ns[:, ibatch] = nsp
            Whs[:, ibatch] = Wheights.astype(np.int32)
        else:
            logger.warning('Data batch #%d only had %d spikes.', ibatch, uproj.shape[1])

        i0 = i0 + Nfilt

    # anothr one of these Params variables transporting parameters to the C++ code
    Params = [1, NrankPC, Nfilt, 0, W.shape[0], 0, NchanNear, Nchan]
    # the total number of templates is the number of templates per batch times the number of batch
    Params[0] = Ws.shape[2] * Ws.shape[3]

    # initialize dissimilarity matrix
    ccb = cp.zeros((nBatches, nBatches), dtype=np.float32, order='F')

    for ibatch in tqdm(range(nBatches), desc="Computing distances"):
        # for every batch, compute in parallel its dissimilarity to ALL other batches
        Wh0 = Whs[:, ibatch]  # this one is the primary batch
        W0 = Ws[..., ibatch]
        mu = mus[..., ibatch]

        # embed the templates from the primary batch back into a full, sparse representation
        W = cp.zeros((nPCs, Nchan, Nfilt), dtype=np.float32, order='F')
        for t in range(Nfilt):
            W[:, iC[:, Wh0[t]], t] = cp.atleast_3d(Ws[:, :, t, ibatch])

        # pairs of templates that live on the same channels are potential "matches"
        # TODO: move_to_config? is the 0.1 here important?
        iMatch = cp.min(cp.abs(iC - Wh0.reshape((1, 1, -1), order='F')), axis=0) < .1

        # compute dissimilarities for iMatch = 1
        iclust, ds = mexDistances2(Params, Ws, W, iMatch, iC, Whs, mus, mu)

        # ds are squared Euclidian distances
        ds = ds.reshape((Nfilt, -1), order='F')  # this should just be an Nfilt-long vector
        ds = cp.maximum(0, ds)

        # weigh the distances according to number of spikes in cluster
        ccb[ibatch, :] = cp.mean(cp.sqrt(ds) * ns, axis=0) / cp.mean(ns, axis=0)

    # ccb = cp.asnumpy(ccb)
    # some normalization steps are needed: zscoring, and symmetrizing ccb
    ccb0 = zscore(ccb, axis=0)
    ccb0 = ccb0 + ccb0.T

    # sort by manifold embedding algorithm
    # iorig is the sorting of the batches
    # ccbsort is the resorted matrix (useful for diagnosing drift)
    ccbsort, iorig = sortBatches2(ccb0)

    logger.info("Finished clustering.")

    if sanity_plots:
        assert plot_widgets is not None, "if sanity_plots is set, then plot_widgets cannot be None"
        plot_dissimilarity_matrices(ccb0, ccbsort, plot_widgets[plot_pos])

    return Bunch(iorig=iorig, ccb0=ccb0, ccbsort=ccbsort)



from contextlib import redirect_stderr
import ctypes
import io
from functools import wraps
import logging
from math import ceil
from textwrap import dedent

import numpy as np
from scipy import signal as ss
import cupy as cp

# Cupy changed core module to _core for some versions but not all
try:
    import cupy._core as cp_core
except ImportError:
    import cupy.core as cp_core

logger = logging.getLogger(__name__)


# LTI filter on GPU (NOTE: inefficient and soon to be deprecated)
# ---------------------------------------------------------------

# TODO: design - Let's seperate out the CUDA *tools* from computations that happen
#              - to rely on CUDA.
def make_kernel(kernel, name, **const_arrs):
    """Compile a kernel and pass optional constant ararys."""
    mod = cp_core.core.compile_with_cache(kernel, prepend_cupy_headers=False)
    b = cp_core.core.memory_module.BaseMemory()
    # Pass constant arrays.
    for n, arr in const_arrs.items():
        b.ptr = mod.get_global_var(n)
        p = cp_core.core.memory_module.MemoryPointer(b, 0)
        p.copy_from_host(arr.ctypes.data_as(ctypes.c_void_p), arr.nbytes)
    return mod.get_function(name)


def get_lfilter_kernel(N, isfortran, reverse=False):
    order = 'f' if isfortran else 'c'
    return dedent("""
    const int N = %d;
    __constant__ float a[N + 1];
    __constant__ float b[N + 1];


    __device__ int get_idx_f(int n, int col, int n_samples, int n_channels) {
        return n_samples * col + n;  // Fortran order.
    }
    __device__ int get_idx_c(int n, int col, int n_samples, int n_channels) {
        return n * n_channels + col;  // C order.
    }

    // LTI IIR filter implemented using a difference equation.
    // see https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.lfilter.html
    extern "C" __global__ void lfilter(
            const float* x, float* y, const int n_samples, const int n_channels){
        // Initialize the state variables.
        float d[N + 1];
        for (int k = 0; k <= N; k++) {
            d[k] = 0.0;
        }

        float xn = 0.0;
        float yn = 0.0;

        int idx = 0;

        // Column index.
        int col = blockIdx.x * blockDim.x + threadIdx.x;


        // IMPORTANT: avoid out of bounds memory accesses, which cause no errors but weird bugs.
        if (col >= n_channels) return;

        for (int n = 0; n < n_samples; n++) {
            idx = get_idx_%s(%s, col, n_samples, n_channels);
            // Load the input element.
            xn = x[idx];
            // Compute the output element.
            yn = (b[0] * xn + d[0]) / a[0];
            // Update the state variables.
            for (int k = 0; k < N; k++) {
                d[k] = b[k + 1] * xn - a[k + 1] * yn + d[k + 1];
            }
            // Update the output array.
            y[idx] = yn;
        }
    }
    """ % (N, order, 'n' if not reverse else 'n_samples - 1 - n'))


def _get_lfilter_fun(b, a, is_fortran=True, axis=0, reverse=False):
    assert axis == 0, "Only filtering along the first axis is currently supported."

    b = np.atleast_1d(b).astype(np.float32)
    a = np.atleast_1d(a).astype(np.float32)
    N = max(len(b), len(a))
    if len(b) < N:
        b = np.pad(b, (0, (N - len(b))), mode='constant')
    if len(a) < N:
        a = np.pad(a, (0, (N - len(a))), mode='constant')
    assert len(a) == len(b)
    kernel = get_lfilter_kernel(N - 1, is_fortran, reverse=reverse)

    lfilter = make_kernel(kernel, 'lfilter', b=b, a=a)

    return lfilter


def _apply_lfilter(lfilter_fun, arr):
    assert isinstance(arr, cp.ndarray)
    if arr.ndim == 1:
        arr = arr[:, np.newaxis]
    n_samples, n_channels = arr.shape

    block = (min(128, n_channels),)
    grid = (int(ceil(n_channels / float(block[0]))),)

    arr = cp.asarray(arr, dtype=np.float32)
    y = cp.zeros_like(arr, order='F' if arr.flags.f_contiguous else 'C', dtype=arr.dtype)

    assert arr.dtype == np.float32
    assert y.dtype == np.float32
    assert arr.shape == y.shape

    lfilter_fun(grid, block, (arr, y, int(y.shape[0]), int(y.shape[1])))
    return y


def lfilter(b, a, arr, axis=0, reverse=False):
    """Perform a linear filter along the first axis on a GPU array."""
    lfilter_fun = _get_lfilter_fun(
        b, a, is_fortran=arr.flags.f_contiguous, axis=axis, reverse=reverse)
    return _apply_lfilter(lfilter_fun, arr)


# GPU FFT-based convolution
# -------------------------

def _clip(x, a, b):
    return max(a, min(b, x))


def pad(fcn_convolve):
    @wraps(fcn_convolve)
    def function_wrapper(x, b, axis=0, **kwargs):
        # add the padding to the array
        xsize = x.shape[axis]
        if 'pad' in kwargs and kwargs['pad']:
            npad = b.shape[axis] // 2
            padd = cp.take(x, cp.arange(npad), axis=axis) * 0
            if kwargs['pad'] == 'zeros':
                x = cp.concatenate((padd, x, padd), axis=axis)
            if kwargs['pad'] == 'constant':
                x = cp.concatenate((padd * 0 + cp.mean(x[:npad]), x, padd + cp.mean(x[-npad:])),
                                   axis=axis)
            if kwargs['pad'] == 'flip':
                pad_in = cp.flip(cp.take(x, cp.arange(1, npad + 1), axis=axis), axis=axis)
                pad_out = cp.flip(cp.take(x, cp.arange(xsize - npad - 1, xsize - 1),
                                          axis=axis), axis=axis)
                x = cp.concatenate((pad_in, x, pad_out), axis=axis)
        # run the convolution
        y = fcn_convolve(x, b, **kwargs)
        # remove padding from both arrays (necessary for x ?)
        if 'pad' in kwargs and kwargs['pad']:
            # remove the padding
            y = cp.take(y, cp.arange(npad, x.shape[axis] - npad), axis=axis)
            x = cp.take(x, cp.arange(npad, x.shape[axis] - npad), axis=axis)
            assert xsize == x.shape[axis]
            assert xsize == y.shape[axis]
        return y
    return function_wrapper


def convolve_cpu(x, b):
    """CPU convolution based on scipy.signal."""
    x = np.asarray(x)
    b = np.asarray(b)
    if b.ndim == 1:
        b = b[:, np.newaxis]
    assert b.ndim == 2
    y = ss.convolve(x, b, mode='same')
    return y


@pad
def convolve_gpu_direct(x, b, **kwargs):
    """Straight GPU FFT-based convolution that fits in memory."""
    if not isinstance(x, cp.ndarray):
        x = np.asarray(x)
    if not isinstance(x, cp.ndarray):
        b = np.asarray(b)
    assert b.ndim == 1
    n = x.shape[0]
    xf = cp.fft.rfft(x, axis=0, n=n)
    if xf.shape[0] > b.shape[0]:
        bp = cp.pad(b, (0, n - b.shape[0]), mode='constant')
        bp = cp.roll(bp, - b.size // 2 + 1)
    else:
        bp = b
    bf = cp.fft.rfft(bp, n=n)[:, np.newaxis]
    y = cp.fft.irfft(xf * bf, axis=0, n=n)
    return y


DEFAULT_CONV_CHUNK = 10_000


@pad
def convolve_gpu_chunked(x, b, pad='flip', nwin=DEFAULT_CONV_CHUNK, ntap=500, overlap=2000):
    """Chunked GPU FFT-based convolution for large arrays.

    This memory-controlled version splits the signal into chunks of n samples.
    Each chunk is tapered in and out, the overlap is designed to get clear of the taper
    splicing of overlaping chunks is done in a cosine way.

    param: pad None, 'zeros', 'constant', 'flip'

    """
    x = cp.asarray(x)
    b = cp.asarray(b)
    assert b.ndim == 1
    n = x.shape[0]
    assert overlap >= 2 * ntap
    # create variables, the gain is to control the splicing
    y = cp.zeros_like(x)
    gain = cp.zeros(n)
    # compute tapers/constants outside of the loop
    taper_in = (-cp.cos(cp.linspace(0, 1, ntap) * cp.pi) / 2 + 0.5)[:, cp.newaxis]
    taper_out = cp.flipud(taper_in)
    assert b.shape[0] < nwin < n
    # this is the convolution wavelet that we shift to be 0 lag
    bp = cp.pad(b, (0, nwin - b.shape[0]), mode='constant')
    bp = cp.roll(bp, - b.size // 2 + 1)
    bp = cp.fft.rfft(bp, n=nwin)[:, cp.newaxis]
    # this is used to splice windows together: cosine taper. The reversed taper is complementary
    scale = cp.minimum(cp.maximum(0, cp.linspace(-0.5, 1.5, overlap - 2 * ntap)), 1)
    splice = (-cp.cos(scale * cp.pi) / 2 + 0.5)[:, cp.newaxis]
    # loop over the signal by chunks and apply convolution in frequency domain
    first = 0
    while True:
        first = min(n - nwin, first)
        last = min(first + nwin, n)
        # the convolution
        x_ = cp.copy(x[first:last, :])
        x_[:ntap] *= taper_in
        x_[-ntap:] *= taper_out
        x_ = cp.fft.irfft(cp.fft.rfft(x_, axis=0, n=nwin) * bp, axis=0, n=nwin)
        # this is to check the gain of summing the windows
        tt = cp.ones(nwin)
        tt[:ntap] *= taper_in[:, 0]
        tt[-ntap:] *= taper_out[:, 0]
        # the full overlap is outside of the tapers: we apply a cosine splicing to this part only
        if first > 0:
            full_overlap_first = first + ntap
            full_overlap_last = first + overlap - ntap
            gain[full_overlap_first:full_overlap_last] *= (1. - splice[:, 0])
            gain[full_overlap_first:full_overlap_last] += tt[ntap:overlap - ntap] * splice[:, 0]
            gain[full_overlap_last:last] = tt[overlap - ntap:]
            y[full_overlap_first:full_overlap_last] *= (1. - splice)
            y[full_overlap_first:full_overlap_last] += x_[ntap:overlap - ntap] * splice
            y[full_overlap_last:last] = x_[overlap - ntap:]
        else:
            y[first:last, :] = x_
            gain[first:last] = tt
        if last == n:
            break
        first += nwin - overlap
    return y


def convolve_gpu(x, b, **kwargs):
    n = x.shape[0]
    # Default chunk size : N samples along the first axis, the one to be chunked and over
    # which to compute the convolution.
    nwin = kwargs.get('nwin', DEFAULT_CONV_CHUNK)
    assert nwin >= 0
    if n <= nwin or nwin == 0:
        return convolve_gpu_direct(x, b)
    else:
        nwin = max(nwin, b.shape[0] + 1)
        return convolve_gpu_chunked(x, b, **kwargs)


def svdecon(X, nPC0=None):
    """
    Input:
    X : m x n matrix

    Output:
    X = U*S*V'

    Description:

    Does equivalent to svd(X,'econ') but faster

        Vipin Vijayan (2014)

    """

    m, n = X.shape

    nPC = nPC0 or min(m, n)

    if m <= n:
        C = cp.dot(X, X.T)
        D, U = cp.linalg.eigh(C, 'U')

        ix = cp.argsort(np.abs(D))[::-1]
        d = D[ix]
        U = U[:, ix]
        d = d[:nPC]
        U = U[:, :nPC]

        V = cp.dot(X.T, U)
        s = cp.sqrt(d)
        V = V / s.T
        S = cp.diag(s)
    else:
        C = cp.dot(X.T, X)
        D, V = cp.linalg.eigh(C)

        ix = cp.argsort(cp.abs(D))[::-1]
        d = D[ix]
        V = V[:, ix]

        # convert evecs from X'*X to X*X'. the evals are the same.
        U = cp.dot(X, V)
        s = cp.sqrt(d)
        U = U / s.T
        S = cp.diag(s)

    return U, S, V


def svdecon_cpu(X):
    U, S, V = np.linalg.svd(cp.asnumpy(X))
    return U, np.diag(S), V


def free_gpu_memory():
    mempool = cp.get_default_memory_pool()
    pinned_mempool = cp.get_default_pinned_memory_pool()
    mempool.free_all_blocks()
    pinned_mempool.free_all_blocks()


# Work around CuPy bugs and limitations
# -----------------------------------------------------------------------------

def mean(x, axis=0):
    if x.ndim == 1:
        return cp.mean(x) if x.size else cp.nan
    else:
        s = list(x.shape)
        del s[axis]
        return (
            cp.mean(x, axis=axis) if x.shape[axis] > 0
            else cp.zeros(s, dtype=x.dtype, order='F'))


def median(a, axis=0):
    """Compute the median of a CuPy array on the GPU."""
    a = cp.asarray(a)

    if axis is None:
        sz = a.size
    else:
        sz = a.shape[axis]
    if sz % 2 == 0:
        szh = sz // 2
        kth = [szh - 1, szh]
    else:
        kth = [(sz - 1) // 2]

    part = cp.partition(a, kth, axis=axis)

    if part.shape == ():
        # make 0-D arrays work
        return part.item()
    if axis is None:
        axis = 0

    indexer = [slice(None)] * part.ndim
    index = part.shape[axis] // 2
    if part.shape[axis] % 2 == 1:
        # index with slice to allow mean (below) to work
        indexer[axis] = slice(index, index + 1)
    else:
        indexer[axis] = slice(index - 1, index + 1)

    return cp.mean(part[tuple(indexer)], axis=axis)


def var(x):
    return cp.var(x, ddof=1) if x.size > 0 else cp.nan


def ones(shape, dtype=None, order=None):
    # HACK: cp.ones() has no order kwarg at the moment !
    x = cp.zeros(shape, dtype=dtype, order=order)
    x.fill(1)
    return x


def zscore(a, axis=0):
    mns = a.mean(axis=axis)
    sstd = a.std(axis=axis, ddof=0)
    return (a - mns) / sstd



import numpy as np

from ..utils import Bunch


""" Some standard probe geometries - please check before using! """

def np1_probe():
    """ Returns a Neuropixels 1 probe as a Bunch object for use in pykilosort """
    probe = Bunch()
    probe.NchanTOT = 385
    probe.chanMap = np.arange(384)
    probe.xc = np.tile(np.array([43., 11., 59., 27.]), 96)
    probe.yc = np.repeat(np.arange(20, 3841, 20.), 2)
    probe.kcoords = np.zeros(384)
    return probe


def np2_probe():
    """ Returns a Neuropixels 2 probe as a Bunch object for use in pykilosort """
    probe = Bunch()
    probe.NchanTOT = 385
    probe.chanMap = np.arange(384)
    probe.xc = np.tile(np.array([0., 32.]), 192)
    probe.yc = np.repeat(np.arange(0, 2866, 15.), 2)
    probe.kcoords = np.zeros(384)
    return probe


def get_4shank_channels_np2(shank):
    """
    Returns the channel indices for a given shank on a np2 4 shank probe
    :param shank: Shank to return, int between 0 and 3 inclusive
    :return: Numpy array
    """
    assert type(shank) == int, 'Shank index must be an integer'
    assert 0 <= shank <= 3, 'Shank index must be between 0 and 3'

    if shank in [0, 1]:
        return np.concatenate((np.arange(48*shank, 48*(shank+1)),
                                        np.arange(48*(shank+2), 48*(shank+3))))
    if shank in [2, 3]:
        return np.concatenate((np.arange(48*(shank+2), 48*(shank+3)),
                                        np.arange(48*(shank+4), 48*(shank+5))))

def np2_4shank_probe(shank=None):
    """
    Returns a Neuropixels 2 4-shank probe as a Bunch object
    :param shank: Optional, return only a single shank, int between 0 and 3 inclusive
    :return: Bunch object for pykilosort
    """
    if shank is not None:
        assert type(shank) == int, 'Shank index must be an integer'
        assert 0 <= shank <= 3, 'Shank index must be between 0 and 3'

    probe = Bunch()
    probe.NchanTOT = 385

    if shank is None:
        # Return whole probe
        probe.chanMap = np.arange(384)
        probe.kcoords = np.zeros(384)
        probe.xc = np.zeros(384)
        probe.yc = np.zeros(384)

        for shank_id in range(4):
            shank_channels = get_4shank_channels_np2(shank_id)
            probe.xc[shank_channels] = np.tile([0., 32.], 48) + shank_id * 200
            probe.yc[shank_channels] = np.repeat(np.arange(2880, 3586, 15.), 2)

        return probe

    if shank in [0, 1]:
        probe.chanMap = np.concatenate((np.arange(48*shank, 48*(shank+1)),
                                        np.arange(48*(shank+2), 48*(shank+3))))
    if shank in [2, 3]:
        probe.chanMap = np.concatenate((np.arange(48*(shank+2), 48*(shank+3)),
                                        np.arange(48*(shank+4), 48*(shank+5))))

    probe.xc = np.tile([0., 32.], 48) + shank * 200
    probe.yc = np.repeat(np.arange(2880, 3586, 15.), 2)
    probe.kcoords = np.zeros(96)

    return probe



from pathlib import Path
import datetime
import json
import logging
import shutil

import numpy as np

import spikeglx
import neuropixel
from ibllib.ephys import spikes
from one.alf.files import get_session_path
from one.remote import aws
from pykilosort import add_default_handler, run, Bunch, __version__
from pykilosort.params import KilosortParams


_logger = logging.getLogger("pykilosort")


def _get_multi_parts_records(bin_file):
    """ Looks for the multiple parts of the recording using sequence files from ibllib"""
    # if multiple files are already provided, do not look for sequence files
    if isinstance(bin_file, list) or isinstance(bin_file, tuple):
        for bf in bin_file:
            if not Path(bf).exists():
                raise FileNotFoundError(bf)
        return bin_file
    # if there is no sequence file attached to the binary file, return just the bin file
    bin_file = Path(bin_file)
    sequence_file = bin_file.parent.joinpath(f"{bin_file.stem.replace('.ap', '.sequence.json')}")
    if not sequence_file.exists():
        if not Path(bin_file).exists():
            raise FileNotFoundError(bin_file)
        else:
            return bin_file
    # if there is a sequence file, return all files if they're all present and this is the first index
    with sequence_file.open() as fid:
        seq = json.load(fid)
    if seq['index'] > 0:
        _logger.warning(f"Multi-part raw ephys: returns empty as this is not the first "
                        f"index in the sequence. Check: {sequence_file}")
        return
    # the common anchor path to look for other meta files is the subject path
    subject_folder = get_session_path(bin_file)
    subject_folder_seq = get_session_path(Path(seq['files'][0])).parents[1]
    # reconstruct path of each binary file, exit with None if one is not found
    cbin_files = []
    for f in seq['files']:
        meta_file = subject_folder.joinpath(Path(f).relative_to(subject_folder_seq))
        cbin_file = next(meta_file.parent.glob(meta_file.stem + '.*bin'), None)
        if cbin_file is None:
            _logger.error(f"Multi-part raw ephys error: missing bin file in folder {meta_file.parent}")
            return
        cbin_files.append(cbin_file)
    return cbin_files


def _sample2v(ap_file):
    md = spikeglx.read_meta_data(ap_file.with_suffix(".meta"))
    s2v = spikeglx._conversion_sample2v_from_meta(md)
    return s2v["ap"][0]


def run_spike_sorting_ibl(bin_file, scratch_dir=None, delete=True,
                          ks_output_dir=None, alf_path=None, log_level='INFO', params=None):
    """
    This runs the spike sorting and outputs the raw pykilosort without ALF conversion
    :param bin_file: binary file full path
    :param scratch_dir: working directory (home of the .kilosort folder) SSD drive preferred.
    :param delete: bool, optional, defaults to True: whether or not to delete the .kilosort temp folder
    :param ks_output_dir: string or Path: output directory defaults to None, in which case it will output in the
     scratch directory.
    :param alf_path: strint or Path, optional: if specified, performs ks to ALF conversion in the specified folder
    :param log_level: string, optional, defaults to 'INFO'
    :return:
    """
    START_TIME = datetime.datetime.now()
    # handles all the paths infrastructure
    assert scratch_dir is not None
    bin_file = _get_multi_parts_records(bin_file)
    scratch_dir.mkdir(exist_ok=True, parents=True)
    ks_output_dir = Path(ks_output_dir) if ks_output_dir is not None else scratch_dir.joinpath('output')
    log_file = scratch_dir.joinpath(f"_{START_TIME.isoformat()}_kilosort.log")
    add_default_handler(level=log_level)
    add_default_handler(level=log_level, filename=log_file)
    session_scratch_dir = scratch_dir.joinpath('.kilosort', bin_file.stem)
    # construct the probe geometry information
    if params is None:
        params = ibl_pykilosort_params(bin_file)
    try:
        _logger.info(f"Starting Pykilosort version {__version__}")
        _logger.info(f"Scratch dir {scratch_dir}")
        _logger.info(f"Output dir {ks_output_dir}")
        _logger.info(f"Data dir {bin_file.parent}")
        _logger.info(f"Log file {log_file}")
        _logger.info(f"Loaded probe geometry for NP{params['probe']['neuropixel_version']}")

        run(bin_file, dir_path=scratch_dir, output_dir=ks_output_dir, **params)
        # move back the QC files to the original probe folder for registration
        for qc_file in session_scratch_dir.rglob('_iblqc_*'):
            shutil.copy(qc_file, ks_output_dir.joinpath(qc_file.name))
        if delete:
            shutil.rmtree(scratch_dir.joinpath(".kilosort"), ignore_errors=True)
    except Exception as e:
        _logger.exception("Error in the main loop")
        raise e
    [_logger.removeHandler(h) for h in _logger.handlers]
    # move the log file and all qcs to the output folder
    shutil.move(log_file, ks_output_dir.joinpath('spike_sorting_pykilosort.log'))

    # convert the pykilosort output to ALF IBL format
    if alf_path is not None:
        s2v = _sample2v(bin_file)
        alf_path.mkdir(exist_ok=True, parents=True)
        spikes.ks2_to_alf(ks_output_dir, bin_file, alf_path, ampfactor=s2v)


def ibl_pykilosort_params(bin_file):
    params = KilosortParams()
    params.preprocessing_function = 'destriping'
    params.probe = probe_geometry(bin_file)
    params.minFR = 0
    params.minfr_goodchannels = 0
    # params = {k: dict(params)[k] for k in sorted(dict(params))}
    return dict(params)


def probe_geometry(bin_file):
    """
    Loads the geometry from the meta-data file of the spikeglx acquisition system
    sr: ibllib.io.spikeglx.Reader or integer with neuropixel version 1 or 2
    """
    if isinstance(bin_file, list):
        sr = spikeglx.Reader(bin_file[0])
        h, ver, s2v = (sr.geometry, sr.major_version, sr.sample2volts[0])
    elif isinstance(bin_file, str) or isinstance(bin_file, Path):
        sr = spikeglx.Reader(bin_file)
        h, ver, s2v = (sr.geometry, sr.major_version, sr.sample2volts[0])
    else:
        print(bin_file)
        assert(bin_file == 1 or bin_file == 2)
        h, ver, s2v = (neuropixel.trace_header(version=bin_file), bin_file, 2.34375e-06)
    nc = h['x'].size
    probe = Bunch()
    probe.NchanTOT = nc + 1
    probe.chanMap = np.arange(nc)
    probe.xc = h['x'] + h['shank'] * 200
    probe.yc = h['y']
    probe.x = h['x']
    probe.y = h['y']
    probe.shank = h['shank']
    probe.kcoords = np.zeros(nc)
    probe.channel_labels = np.zeros(nc, dtype=int)
    probe.sample_shift = h['sample_shift']
    probe.h, probe.neuropixel_version, probe.sample2volt = (h, ver, s2v)
    return probe


def download_test_data(local_folder):
    return aws.s3_download_folder('spikesorting/integration_tests', local_folder)


def download_benchmark_data(local_folder):
    return aws.s3_download_folder('spikesorting/benchmarks', local_folder)



import logging
import os
from math import sqrt, ceil

import numpy as np
import cupy as cp
from tqdm.auto import tqdm

from pykilosort.cptools import svdecon, svdecon_cpu, median, free_gpu_memory, ones
from pykilosort.cluster import isolated_peaks_new, get_SpikeSample, getClosestChannels
from pykilosort.utils import Bunch, get_cuda, _extend, LargeArrayWriter, plot_diagnostics

logger = logging.getLogger(__name__)


def extractTemplatesfromSnippets(
    data_loader=None, probe=None, params=None, Nbatch=None, nPCs=None
):
    # this function is very similar to extractPCfromSnippets.
    # outputs not just the PC waveforms, but also the template "prototype",
    # basically k-means clustering of 1D waveforms.

    NT = params.NT
    # skip every this many batches
    nskip = params.nskip
    nPCs = nPCs or params.nPCs
    nt0min = params.nt0min

    k = 0
    # preallocate matrix to hold 1D spike snippets
    # dd = cp.zeros((params.nt0, int(5e4)), dtype=np.float32, order='F')
    dds = []

    for ibatch in tqdm(range(0, Nbatch, nskip), desc="Extracting templates"):
        # load data to GPU
        dataRAW = data_loader.load_batch(ibatch)

        # find isolated spikes from each batch
        row, col, mu = isolated_peaks_new(dataRAW, params)

        # for each peak, get the voltage snippet from that channel
        c = get_SpikeSample(dataRAW, row, col, params)

        # if k + c.shape[1] > dd.shape[1]:
        #     dd = cp.pad(dd, (0, dd.shape[1]), mode='constant')

        # dd[:, k:k + c.shape[1]] = c
        dds.append(c)
        k = k + c.shape[1]
        if k > 1e5:
            break

    # discard empty samples
    # dd = dd[:, :k]
    dd = cp.asfortranarray(cp.concatenate(dds, axis=1).astype(np.float32))

    # initialize the template clustering with random waveforms
    uu = np.random.permutation(dd.shape[1])[:nPCs]
    wTEMP = dd[:, uu]
    wTEMP = wTEMP / cp.sum(wTEMP ** 2, axis=0) ** 0.5  # normalize them

    for i in range(10):
        # at each iteration, assign the waveform to its most correlated cluster
        cc = cp.dot(wTEMP.T, dd)
        imax = cp.argmax(cc, axis=0)
        amax = cc[imax, np.arange(cc.shape[1])]
        for j in range(nPCs):
            # weighted average to get new cluster means
            wTEMP[:, j] = cp.dot(dd[:, imax == j], amax[imax == j].T)
        wTEMP = wTEMP / cp.sum(wTEMP ** 2, axis=0) ** 0.5  # unit normalize

    # the PCs are just the left singular vectors of the waveforms
    U, Sv, V = svdecon(dd)

    # take as many as needed
    wPCA = U[:, :nPCs]

    # adjust the arbitrary sign of the first PC so its negativity is downward
    wPCA[:, 0] = -wPCA[:, 0] * cp.sign(wPCA[nt0min, 0])

    return wTEMP, wPCA


# TODO: design - if we just need a couple of params then we can just pass those in.
def getKernels(params):
    # this function makes upsampling kernels for the temporal components.
    # those are used for interpolating the biggest negative peak,
    # and aligning the template to that peak with sub-sample resolution
    # needs nup, the interpolation factor (default = 10)
    # also needs sig, the interpolation smoothness (default = 1)

    nup = params.nup
    sig = params.sig

    nt0min = params.nt0min
    nt0 = params.nt0

    xs = cp.arange(1, nt0 + 1)
    ys = cp.linspace(0.5, nt0 + 0.5, nt0 * nup + 1)[:-1]

    # these kernels are just standard kriging interpolators

    # first compute distances between the sample coordinates
    # for some reason, this seems to be circular, although the waveforms are not circular
    # I think the reason had to do with some constant offsets in some channels?
    d = cp.mod(xs[:, np.newaxis] - xs[np.newaxis, :] + nt0, nt0)
    d = cp.minimum(d, nt0 - d)
    # the kernel covariance uses a squared exponential of spatial scale sig
    Kxx = cp.exp(-(d ** 2) / sig ** 2)

    # do the same for the kernel similarities between upsampled "test" timepoints and
    # the original coordinates
    d = cp.mod(ys[:, np.newaxis] - xs[np.newaxis, :] + nt0, nt0)
    d = cp.minimum(d, nt0 - d)
    Kyx = cp.exp(-(d ** 2) / sig ** 2)

    # the upsampling matrix is given by the following formula,
    # with some light diagonal regularization of the matrix inversion
    B = cp.dot(Kyx, cp.linalg.inv(Kxx + 0.01 * cp.eye(nt0)))
    B = B.reshape((nup, nt0, nt0), order="F")

    # A is just a slice through this upsampling matrix corresponding to the most negative point
    # this is used to compute the biggest negative deflection (after upsampling)
    A = cp.squeeze(B[:, nt0min - 1, :])
    B = cp.transpose(B, [1, 2, 0])

    return A.astype(np.float64), B.astype(np.float64)


def getMeUtU(iU, iC, mask, Nnearest, Nchan):
    # function [UtU, maskU, iList] = getMeUtU(iU, iC, mask, Nnearest, Nchan)
    # this function determines if two templates share any channels
    # iU are the channels that each template is assigned to, one main channel per template
    # iC has as column K the list of neigboring channels for channel K
    # mask are the weights assigned for the corresponding neighboring channels
    # in iC (gaussian-decaying)

    Nfilt = iU.size

    # create a sparse matrix with ones if a channel K belongs to a template
    U = np.zeros((Nchan, Nfilt), dtype=np.float32, order="F")

    # use the template primary channel to obtain its neighboring channels from iC
    ix = cp.asnumpy(iC[:, iU]).squeeze() + np.arange(0, Nchan * Nfilt, Nchan).astype(np.int32)

    # WARNING: transpose because the indexing assumes F ordering.
    U.T.flat[ix] = 1  # use this as an awkward index into U

    # if this is 0, the templates had not pair of channels in common
    UtU = (cp.dot(U.T, U) > 0).astype(np.int32)

    # we also return the masks for each template, picked from the corresponding mask of
    # their primary channel
    maskU = mask[:, iU]

    # # sort template pairs in order of how many channels they share
    # isort = cp.argsort(UtU, axis=0)[::-1]
    # iList = isort[:Nnearest, :]  # take the Nnearest templates for each template

    return UtU, maskU  # , iList


def getMeWtW2(W, U0, Nnearest=None):
    # this function compute the correlation between any two pairs of templates

    # it relies on the fact that the W and U0 are unit normalized, so that the product of a
    # template with itself is 1, as it should be if we're trying to calculate correlations

    # takes input the temporal and spatial factors of the low-rank template, as
    # well as the number of most similar template pairs desired to be output in
    # iList

    nt0, Nfilt, Nrank = W.shape
    WtW = cp.zeros((Nfilt, Nfilt), dtype=np.float32, order="F")

    # since the templates are factorized into orthonormal components, we can compute dot products
    # one dimension at a time
    for i in range(Nrank):
        for j in range(Nrank):
            #  this computes the spatial dot product
            utu0 = cp.dot(U0[:, :, i].T, U0[:, :, j])
            #  this computes the temporal dot product
            wtw0 = cp.dot(W[:, :, i].T, W[:, :, j])

            # the element-wise product of these is added to the matrix of correlatioons
            WtW = WtW + wtw0 * utu0

    # also return a list of most correlated template pairs
    isort = cp.argsort(WtW, axis=0)[::-1]

    if Nnearest:
        # if we don't have enough templates yet, just wrap the indices around the range 1:Nfilt
        iNear = cp.mod(cp.arange(Nnearest), Nfilt)
        iList = isort[iNear, :]  # return the list of pairs for each template
        return WtW, iList
    else:
        return WtW


def custom_lexsort(arrays):
    """
    Lexsort of 1D Cupy arrays, last array given is used for the primary sort order, second-last
    for secondary sort order and so on
    :param arrays: List of 1D Cupy arrays, all lengths must match
    :return: Cupy array of indices that sort the arrays according to the above
    """
    # Check arrays are one-dimensional and have equal length
    for array in arrays:
        assert array.ndim == 1
    n = arrays[0].shape[0]
    for array in arrays[1:]:
        assert array.shape[0] == n

    # Concatenate arrays and pass to cupy lexsort function
    lex_array = cp.concatenate([array.reshape(1, -1) for array in arrays], axis=0)
    return cp.lexsort(lex_array)


def mexGetSpikes2(Params, drez, wTEMP, iC):
    code, constants = get_cuda("mexGetSpikes2")

    NT = int(Params[0])
    Nchan = int(Params[9])
    nt0 = int(Params[4])
    # Nnearest = int(Params[5])
    Nrank = int(Params[14])

    maxFR = constants.maxFR
    Nthreads = constants.Nthreads

    # tpB = (8, 2 * nt0 - 1)
    # tpF = (16, Nnearest)
    tpS = (nt0, 16)

    d_Params = cp.asarray(Params, dtype=np.float64, order="F")
    d_data = cp.asarray(drez, dtype=np.float32, order="F")
    d_W = cp.asarray(wTEMP, dtype=np.float32, order="F")
    d_iC = cp.asarray(iC, dtype=np.int32, order="F")

    d_counter = cp.zeros(2, dtype=np.int32, order="F")
    d_dout = cp.zeros((NT, Nchan), dtype=np.float32, order="F")
    d_dfilt = cp.zeros((Nrank, NT, Nchan), dtype=np.float32, order="F")
    d_err = cp.zeros(NT, dtype=np.float32, order="F")
    d_kkmax = cp.zeros((NT, Nchan), dtype=np.int32, order="F")
    d_kk = cp.zeros(NT, dtype=np.int32, order="F")
    d_ftype = cp.zeros(NT, dtype=np.int32, order="F")
    d_st = cp.zeros(maxFR, dtype=np.int32, order="F")
    d_id = cp.zeros(maxFR, dtype=np.int32, order="F")
    d_x = cp.zeros(maxFR, dtype=np.float32, order="F")
    d_st1 = cp.zeros(maxFR, dtype=np.int32, order="F")
    d_id1 = cp.zeros(maxFR, dtype=np.int32, order="F")

    counter = np.zeros(2, dtype=np.int32, order="F")

    # filter the data with the temporal templates
    Conv1D = cp.RawKernel(code, "Conv1D")
    Conv1D((Nchan,), (Nthreads,), (d_Params, d_data, d_W, d_dfilt))

    # sum each template across channels, square, take max
    sumChannels = cp.RawKernel(code, "sumChannels")
    sumChannels(
        (int(NT / Nthreads),), (Nthreads,), (d_Params, d_dfilt, d_dout, d_kkmax, d_iC)
    )

    # compute the best filter
    bestFilter = cp.RawKernel(code, "bestFilter")
    bestFilter(
        (int(NT / Nthreads),),
        (Nthreads,),
        (d_Params, d_dout, d_err, d_ftype, d_kkmax, d_kk),
    )

    # ignore peaks that are smaller than another nearby peak
    cleanup_spikes = cp.RawKernel(code, "cleanup_spikes")
    cleanup_spikes(
        (int(NT / Nthreads),),
        (Nthreads,),
        (d_Params, d_err, d_ftype, d_x, d_st, d_id, d_counter),
    )

    # ignore peaks that are smaller than another nearby peak
    cleanup_heights = cp.RawKernel(code, "cleanup_heights")
    cleanup_heights(
        (1 + int(maxFR // 32),),
        (32,),
        (d_Params, d_x, d_st, d_id, d_st1, d_id1, d_counter),
    )

    # Order of peaks found can vary depending on thread execution times, sort for determinism
    n_spikes = int(min(maxFR, d_counter[1]))
    sorted_ids = custom_lexsort([d_id1[:n_spikes], d_st1[:n_spikes]])
    d_st1[:n_spikes] = d_st1[:n_spikes][sorted_ids]
    d_id1[:n_spikes] = d_id1[:n_spikes][sorted_ids]

    # add new spikes to 2nd counter
    counter[0] = d_counter[1]
    counter[0] = min(maxFR, counter[0])

    d_WU = cp.zeros((nt0, Nchan, counter[0]), dtype=np.float32, order="F")
    # d_WU1 = cp.zeros((nt0, Nchan, counter[0]), dtype=np.float32, order='F')

    # update dWU here by adding back to subbed spikes
    extract_snips = cp.RawKernel(code, "extract_snips")
    extract_snips((Nchan,), tpS, (d_Params, d_st1, d_id1, d_counter, d_data, d_WU))

    # QUESTION: why a copy here??
    # if counter[0] > 0:
    #     d_WU1[...] = d_WU[...]

    del (
        d_ftype,
        d_kkmax,
        d_err,
        d_st,
        d_id,
        d_st1,
        d_x,
        d_kk,
        d_id1,
        d_counter,
        d_Params,
        d_dfilt,
    )
    return d_WU, d_dout


def mexSVDsmall2(Params, dWU, W, iC, iW, Ka, Kb):
    code, constants = get_cuda("mexSVDsmall2")

    Nthreads = constants.Nthreads

    Nfilt = int(Params[1])
    nt0 = int(Params[4])
    Nrank = int(Params[6])
    Nchan = int(Params[9])

    d_Params = cp.asarray(Params, dtype=np.float64, order="F")

    d_dWU = cp.asarray(dWU, dtype=np.float64, order="F")
    d_iC = cp.asarray(iC, dtype=np.int32, order="F")
    d_iW = cp.asarray(iW, dtype=np.int32, order="F")

    d_A = cp.asarray(Ka, dtype=np.float64, order="F")
    d_B = cp.asarray(Kb, dtype=np.float64, order="F")

    d_U = cp.zeros((Nchan, Nfilt, Nrank), dtype=np.float64, order="F")
    d_mu = cp.zeros(Nfilt, dtype=np.float64, order="F")

    d_W = cp.asarray(W, dtype=np.float64, order="F")

    d_wtw = cp.zeros((nt0, nt0, Nfilt), dtype=np.float64, order="F")
    d_dWUb = cp.zeros((nt0, Nchan, Nfilt), dtype=np.float64, order="F")

    tpS = (nt0, int(Nthreads // nt0))
    tpK = (Nrank, int(Nthreads // Nrank))

    blankdWU = cp.RawKernel(code, "blankdWU")
    blankdWU((Nfilt,), tpS, (d_Params, d_dWU, d_iC, d_iW, d_dWUb))

    # compute dWU * dWU'
    getwtw = cp.RawKernel(code, "getwtw")
    getwtw((Nfilt,), tpS, (d_Params, d_dWUb, d_wtw))

    # get W by power svd iterations
    getW = cp.RawKernel(code, "getW")
    getW((Nfilt,), (nt0,), (d_Params, d_wtw, d_W))

    # compute U by W' * dWU
    getU = cp.RawKernel(code, "getU")
    getU((Nfilt,), tpK, (d_Params, d_dWUb, d_W, d_U))

    # normalize U, get S, get mu, renormalize W
    reNormalize = cp.RawKernel(code, "reNormalize")
    reNormalize((Nfilt,), (nt0,), (d_Params, d_A, d_B, d_W, d_U, d_mu))

    del d_wtw, d_Params, d_dWUb

    return d_W, d_U, d_mu


def mexMPnu8(Params, dataRAW, U, W, mu, iC, iW, UtU, iList, wPCA, params):
    code, constants = get_cuda("mexMPnu8")
    maxFR = int(constants.maxFR)
    nmaxiter = int(constants.nmaxiter)
    Nthreads = int(constants.Nthreads)

    NT = int(Params[0]) # NT = (unsigned int) Params[0];
    Nfilt = int(Params[1])
    nt0 = int(Params[4])
    Nnearest = int(Params[5])
    Nrank = int(Params[6])
    NchanU = int(Params[10])
    Nchan = int(Params[9])

    d_Params = cp.asarray(Params, dtype=np.float64, order="F")

    d_draw = cp.asarray(dataRAW, dtype=np.float32, order="F")
    d_U = cp.asarray(U, dtype=np.float32, order="F")
    d_W = cp.asarray(W, dtype=np.float32, order="F")
    d_mu = cp.asarray(mu, dtype=np.float32, order="F")
    d_iC = cp.asarray(iC, dtype=np.int32, order="F")
    d_iW = cp.asarray(iW, dtype=np.int32, order="F")
    d_UtU = cp.asarray(UtU, dtype=bool, order="F")
    d_iList = cp.asarray(iList, dtype=np.int32, order="F")
    d_wPCA = cp.asarray(wPCA, dtype=np.float32, order="F")

    d_nsp = cp.zeros(Nfilt, dtype=np.int32, order="F")
    d_dWU = cp.zeros((nt0, Nchan, Nfilt), dtype=np.float64, order="F")

    d_dout = cp.zeros((2 * NT, Nfilt), dtype=np.float32, order="F")
    d_data = cp.zeros((NT, Nfilt, Nrank), dtype=np.float32, order="F")
    d_err = cp.zeros(NT, dtype=np.float32, order="F")
    d_ftype = cp.zeros(NT, dtype=np.int32, order="F")
    d_eloss = cp.zeros(NT, dtype=np.float32, order="F")
    d_st = cp.zeros(maxFR, dtype=np.int32, order="F")
    d_id = cp.zeros(maxFR, dtype=np.int32, order="F")
    d_x = cp.zeros(maxFR, dtype=np.float32, order="F")
    d_y = cp.zeros(maxFR, dtype=np.float32, order="F")
    d_z = cp.zeros(maxFR, dtype=np.float32, order="F")

    d_counter = cp.zeros(2, dtype=np.int32, order="F")
    d_count = cp.zeros(nmaxiter, dtype=np.int32, order="F")
    d_feat = cp.zeros((Nnearest, maxFR), dtype=np.float32, order="F")
    d_featPC = cp.zeros((NchanU, Nrank, maxFR), dtype=np.float32, order="F")

    d_idx = cp.zeros(maxFR, dtype=np.int32, order="F")

    counter = np.zeros(2, dtype=np.int32, order="F")

    tpF = (16, Nnearest)
    tpS = (nt0, 16)
    tpPC = (NchanU, Nrank)

    # filter the data with the spatial templates
    spaceFilter = cp.RawKernel(code, "spaceFilter")
    spaceFilter((Nfilt,), (Nthreads,), (d_Params, d_draw, d_U, d_iC, d_iW, d_data))

    # filter the data with the temporal templates
    timeFilter = cp.RawKernel(code, "timeFilter")
    timeFilter((Nfilt,), (Nthreads,), (d_Params, d_data, d_W, d_dout))

    # compute the best filter
    bestFilter = cp.RawKernel(code, "bestFilter")
    bestFilter(
        (int(NT // Nthreads),),
        (Nthreads,),
        (d_Params, d_dout, d_mu, d_err, d_eloss, d_ftype),
    )

    if params.stable_mode and not params.deterministic_mode:
        d_draw64 = cp.array(d_draw, dtype=np.float64)

    # loop to find and subtract spikes
    for k in range(int(Params[3])):
        # ignore peaks that are smaller than another nearby peak
        cleanup_spikes = cp.RawKernel(code, "cleanup_spikes")
        cleanup_spikes(
            (int(NT // Nthreads),),
            (Nthreads,),
            (
                d_Params,
                d_dout,
                d_mu,
                d_err,
                d_eloss,
                d_ftype,
                d_st,
                d_id,
                d_x,
                d_y,
                d_z,
                d_counter,
            ),
        )

        # add new spikes to 2nd counter
        counter[:] = cp.asnumpy(d_counter[:])
        if counter[0] > maxFR:
            logger.warning("Firing rate limit hit for batch.")
            counter[0] = maxFR
            d_counter[0] = counter[0]

        # extract template features before subtraction
        if Params[12] > 1:
            extractFEAT = cp.RawKernel(code, "extractFEAT")
            extractFEAT(
                (64,),
                tpF,
                (d_Params, d_st, d_id, d_counter, d_dout, d_iList, d_mu, d_feat),
            )

        if params.deterministic_mode:
            if params.stable_mode:
                d_stSort = d_st[counter[1]:counter[0]] # cudaMemcpy( d_stSort, d_st+counter[1], (counter[0] - counter[1])*sizeof(int), cudaMemcpyDeviceToDevice );
                d_idx[:counter[0]-counter[1]] = cp.argsort(d_stSort) # cdp_simple_quicksort<<< 1, 1 >>>(d_stSort, d_idx, 0, counter[0] - counter[1] - 1, 0);
            else:
                raise ValueError("Stablemode required for deterministic calculations.")
                # This is allowed in the MATLAB version runtime but it doesn't really make sense
                # and isn't recommended so let's not anyone fall into the trap without knowing.
                # d_idx = cp.arange(0, counter[0] - counter[1])

            if Nchan < Nthreads:
                subtract_spikes_v2 = cp.RawKernel(code, "subtract_spikes_v2")
                subtract_spikes_v2(
                    (1,),
                    (Nchan,),
                    (d_Params, d_st, d_idx, d_id, d_y, d_counter, d_draw, d_W, d_U),
                )
            else:
                subtract_spikes_v2 = cp.RawKernel(code, "subtract_spikes_v2")
                subtract_spikes_v2(
                    (Nchan / Nthreads,),
                    (Nthreads,),
                    (d_Params, d_st, d_idx, d_id, d_y, d_counter, d_draw, d_W, d_U),
                )

            spaceFilterUpdate = cp.RawKernel(code, "spaceFilterUpdate")
            spaceFilterUpdate(
                (Nfilt,),
                (2 * nt0 - 1,),
                (
                    d_Params,
                    d_draw,
                    d_U,
                    d_UtU,
                    d_iC,
                    d_iW,
                    d_data,
                    d_st,
                    d_id,
                    d_counter,
                ),
            )
        else:
            if params.stable_mode:
                subtract_spikes_v4 = cp.RawKernel(code, "subtract_spikes_v4")
                subtract_spikes_v4(
                    (Nfilt,),
                    tpS,
                    (d_Params, d_st, d_id, d_y, d_counter, d_draw64, d_W, d_U),
                )

                spaceFilterUpdate_v2 = cp.RawKernel(code, "spaceFilterUpdate_v2")
                spaceFilterUpdate_v2(
                    (Nfilt,),
                    (2 * nt0 - 1,),
                    (
                        d_Params,
                        d_draw64,
                        d_U,
                        d_UtU,
                        d_iC,
                        d_iW,
                        d_data,
                        d_st,
                        d_id,
                        d_counter,
                    ),
                )
            else:
                # subtract spikes from raw data here
                subtract_spikes = cp.RawKernel(code, "subtract_spikes")
                subtract_spikes(
                    (Nfilt,),
                    tpS,
                    (d_Params, d_st, d_id, d_y, d_counter, d_draw, d_W, d_U),
                )

                # filter the data with the spatial templates
                spaceFilterUpdate = cp.RawKernel(code, "spaceFilterUpdate")
                spaceFilterUpdate(
                    (Nfilt,),
                    (2 * nt0 - 1,),
                    (
                        d_Params,
                        d_draw,
                        d_U,
                        d_UtU,
                        d_iC,
                        d_iW,
                        d_data,
                        d_st,
                        d_id,
                        d_counter,
                    ),
                )

        # filter the data with the temporal templates
        timeFilterUpdate = cp.RawKernel(code, "timeFilterUpdate")
        timeFilterUpdate(
            (Nfilt,),
            (2 * nt0 - 1,),
            (d_Params, d_data, d_W, d_UtU, d_dout, d_st, d_id, d_counter),
        )

        if counter[0] - counter[1] > 0:
            bestFilterUpdate = cp.RawKernel(code, "bestFilterUpdate")
            bestFilterUpdate(
                (counter[0] - counter[1],),
                (2 * nt0 - 1,),
                (
                    d_Params,
                    d_dout,
                    d_mu,
                    d_err,
                    d_eloss,
                    d_ftype,
                    d_st,
                    d_id,
                    d_counter,
                ),
            )

        d_count[k + 1] = d_counter[0]

        # update 1st counter from 2nd counter
        d_counter[1] = d_counter[0]

    if params.stable_mode and not params.deterministic_mode:
        d_draw = cp.array(d_draw64, dtype=np.float32)

    # compute PC features from residuals + subtractions
    # TODO: design - let's not use numeric indexing into the Params array. It's much more difficult to read.
    if Params[12] > 0:
        computePCfeatures = cp.RawKernel(code, "computePCfeatures")
        computePCfeatures(
            (Nfilt,),
            tpPC,
            (
                d_Params,
                d_counter,
                d_draw,
                d_st,
                d_id,
                d_y,
                d_W,
                d_U,
                d_mu,
                d_iW,
                d_iC,
                d_wPCA,
                d_featPC,
            ),
        )

    if params.stable_mode:
        # d_idx = array of time sorted indices
        d_idx[:counter[0]] = cp.argsort(d_st[:counter[0]]) # cdp_simple_quicksort<<< 1, 1 >>>(d_stSort, d_idx, 0, counter[0] - counter[1] - 1, 0);
    else:
        d_idx = cp.arange(0, counter[0])

    # update dWU here by adding back to subbed spikes.
    average_snips = cp.RawKernel(code, "average_snips")
    average_snips(
        (Nfilt,),
        tpS,
        (
            d_Params,
            d_st,
            d_idx,
            d_id,
            d_x,
            d_y,
            d_counter,
            d_draw,
            d_W,
            d_U,
            d_dWU,
            d_nsp,
            d_mu,
            d_z,
        ),
    )

    if counter[0] < maxFR:
        minSize = counter[0]
    else:
        minSize = maxFR

    del d_counter, d_Params, d_ftype, d_err, d_eloss, d_z, d_dout, d_data

    # Sort to ensure determinism as different thread execution times leads to spikes being found
    # in a random order

    sorted_idx = custom_lexsort([d_y[:minSize], d_id[:minSize], d_st[:minSize]])

    # Indexing a f-contiguous array returns a c-contiguous array so we need to recast back into
    # fortran order
    return (
        d_st[:minSize][sorted_idx],
        d_id[:minSize][sorted_idx],
        d_y[:minSize][sorted_idx],
        cp.asfortranarray(d_feat[..., :minSize][..., sorted_idx]),
        d_dWU,
        d_draw,
        d_nsp,
        cp.asfortranarray(d_featPC[..., :minSize][..., sorted_idx]),
        d_x[:minSize][sorted_idx],
    )


def mexWtW2(W1, W2, UtU):
    """
    Calculates the correlations between two sets of temporal components at different time lags and
    multiplies these by the correlations of their spatial components to get the template
    correlations at all time lags

    :param W1: Temporal components,
                cupy array with shape (n_times, n_templates)
    :param W2: Temporal components,
                cupy array with shape (n_times, n_templates)
    :param UtU: Correlations between spatial components,
                cupy array with shape (n_templates, n_templates)
    :return: Template correlations across all time lags,
                cupy array with shape (n_templates, n_templates, 2*n_times-1), fortran order
    """

    d_W1 = cp.asarray(W1, dtype=np.float32, order="F")
    d_W2 = cp.asarray(W2, dtype=np.float32, order="F")
    d_UtU = cp.asarray(UtU, dtype=np.float32, order="F")

    code, constants = get_cuda("mexWtW2")
    nblock = constants.nblock

    n_times, n_templates = d_W1.shape

    WtW = cp.zeros((n_templates, n_templates, 2 * n_times - 1), dtype=np.float32, order="F")

    # Setting parameters for CUDA kernel
    grid = (1 + int(n_templates // nblock), 1 + int(n_templates // nblock))
    block = (nblock, nblock)

    # CUDA function that calculates the correlations
    crossFilter = cp.RawKernel(code, "crossFilter")
    crossFilter(grid, block, (d_W1, d_W2, d_UtU, WtW, n_templates, n_times))

    assert cp.sum(cp.isnan(WtW)) == 0

    return WtW


def getMeWtW(W, U0, Nnearest=None):
    # this function computes correlation between templates at ALL timelags from each other
    # takes the max over timelags to obtain a similarity score
    # also returns lists of most similar templates to each template
    # takes as input the low-rank factorization of templates (W for time and U0
    # for space)

    # W is timesamples (default = 61 ), by number of templates, by rank (default = 3)
    nt0, Nfilt, Nrank = W.shape

    Params = [1, Nfilt, 0, 0, 0, 0, 0, 0, 0, nt0]

    # initialize correlation matrix for all timelags
    WtW = cp.zeros((Nfilt, Nfilt, 2 * nt0 - 1), dtype=np.float32, order="F")
    for i in range(Nrank):
        for j in range(Nrank):
            # the dot product factorizes into separable products for each spatio-temporal component
            utu0 = cp.dot(U0[:, :, i].T, U0[:, :, j])  # spatial products
            # temporal convolutions get multiplied with the spatial products
            wtw0 = mexWtW2(W[:, :, i], W[:, :, j], utu0)
            # add it to the full correlation array
            WtW = WtW + wtw0

    # the maximum across timelags accounts for sample alignment mismatch
    cc = cp.max(WtW, axis=2)

    if Nnearest:
        isort = cp.argsort(cc, axis=0)[::-1]
        # if we don't have enough templates yet, just wrap the indices around the range 1:Nfilt
        iNear = cp.mod(cp.arange(Nnearest), Nfilt)
        iList = isort[iNear, :]  # return the list of pairs for each template
        return WtW, iList
    else:
        return WtW


def triageTemplates2(params, iW, C2C, W, U, dWU, mu, nsp, ndrop):

    # This function checks if some templates should be dropped
    # either because they are very similar to another template,
    # or because they are not catching any spikes, (low mean firing rate).
    # Takes as inputs almost all the information that determines templates, and
    # outputs the same variables back after removing some clusters.

    # this is the firing rate threshold
    m0 = params.minFR * params.NT / params.fs
    idrop = nsp < m0  # drop any templates with firing rate below this

    # remove those templates everywhere
    W = cp.asfortranarray(W[:, ~idrop, :])
    U = cp.asfortranarray(U[:, ~idrop, :])
    dWU = cp.asfortranarray(dWU[:, :, ~idrop])
    mu = mu[~idrop]
    nsp = nsp[~idrop]
    # keep track of how many templates have been removed this way
    ndrop[0] = 0.9 * ndrop[0] + 0.1 * idrop.sum()

    # compute pairwise correlations between templates
    cc = getMeWtW2(W, U, None)
    cc = cc - cp.diag(cp.diag(cc))  # exclude the diagonal

    sd = sqrt(10)  # this is hard-coded here

    # compute a score for the separation of the means
    r0 = 4 * sd / cp.abs(mu[:, np.newaxis] - mu[np.newaxis, :])
    # determine which template has more spikes (that one survives)
    rdir = (nsp[:, np.newaxis] - nsp[np.newaxis, :]) < 0
    # for each pair of template, score their similarity by their template correlation,
    # and amplitude separation
    ipair = (cc > 0.9) & (r0 > 1) & rdir
    # for each template, find its most similar other template
    amax = cp.max(ipair, axis=1)
    # if this score is 1, then all the criteria have bene met for dropping this template
    idrop = amax > 0

    # remove these templates everywhere like before
    W = cp.asfortranarray(W[:, ~idrop, :])
    U = cp.asfortranarray(U[:, ~idrop, :])
    dWU = cp.asfortranarray(dWU[:, :, ~idrop])
    mu = mu[~idrop]
    nsp = nsp[~idrop]
    # keep track of how many templates have been removed this way
    ndrop[1] = 0.9 * ndrop[1] + 0.1 * idrop.sum()

    return W, U, dWU, mu, nsp, ndrop


def learnAndSolve8b(ctx, sanity_plots=False, plot_widgets=None, plot_pos=None):
    """This is the main optimization. Takes the longest time and uses the GPU heavily."""

    Nbatch = ctx.intermediate.Nbatch
    params = ctx.params
    probe = ctx.probe
    ir = ctx.intermediate
    data_loader = ir.data_loader

    iorig = ir.iorig

    # TODO: move_to_config
    NrankPC = 6  # this one is the rank of the PCs, used to detect spikes with threshold crossings
    Nrank = 3  # this one is the rank of the templates

    if params.seed is not None:
        np.random.seed(params.seed)

    wTEMP, wPCA = extractTemplatesfromSnippets(
        data_loader=data_loader, probe=probe, params=params, Nbatch=Nbatch, nPCs=NrankPC
    )

    # move these to the GPU
    wPCA = cp.asarray(wPCA[:, :Nrank], dtype=np.float32, order="F")
    wTEMP = cp.asarray(wTEMP, dtype=np.float32, order="F")
    wPCAd = cp.asarray(
        wPCA, dtype=np.float64, order="F"
    )  # convert to double for extra precision

    nt0 = params.nt0
    nt0min = params.nt0min
    nBatches = Nbatch
    NT = params.NT
    Nfilt = params.Nfilt
    Nchan = probe.Nchan

    # two variables for the same thing? number of nearest channels to each primary channel
    # TODO: unclear - let's fix this
    NchanNear = min(probe.Nchan, 32)
    Nnearest = min(probe.Nchan, 32)

    # decay of gaussian spatial mask centered on a channel
    sigmaMask = params.sigmaMask

    # find the closest NchanNear channels, and the masks for those channels
    iC, mask, C2C = getClosestChannels(probe, sigmaMask, NchanNear)

    # batch order schedule is a random permutation of all batches
    ischedule = np.random.permutation(nBatches)
    i1 = np.arange(nBatches)

    irounds = np.concatenate((ischedule, i1))

    niter = irounds.size

    # these two flags are used to keep track of what stage of model fitting we're at
    # flag_final = 0
    flag_resort = 1

    # this is the absolute temporal offset in seconds corresponding to the start of the
    # spike sorted time segment
    t0 = 0  # ceil(params.trange(1) * ops.fs)

    nInnerIter = 60  # this is for SVD for the power iteration

    # schedule of learning rates for the model fitting part
    # starts small and goes high, it corresponds approximately to the number of spikes
    # from the past that were averaged to give rise to the current template
    pmi = np.exp(
        -1.0 / np.linspace(params.momentum[0], params.momentum[1], niter - nBatches)
    )

    Nsum = min(Nchan, 7)  # how many channels to extend out the waveform in mexgetspikes
    # lots of parameters passed into the CUDA scripts
    Params = np.array(
        [
            NT,
            Nfilt,
            params.Th[0],
            nInnerIter,
            nt0,
            Nnearest,
            Nrank,
            params.lam,
            pmi[0],
            Nchan,
            NchanNear,
            params.nt0min,
            1,
            Nsum,
            NrankPC,
            params.Th[0],
        ],
        dtype=np.float64,
    )

    # W0 has to be ordered like this
    W0 = cp.transpose(
        cp.atleast_3d(cp.asarray(wPCA, dtype=np.float64, order="F")), [0, 2, 1]
    )

    # initialize the list of channels each template lives on
    iList = cp.zeros((Nnearest, Nfilt), dtype=np.int32, order="F")

    # initialize average number of spikes per batch for each template
    nsp = cp.zeros((0, 1), dtype=np.float64, order="F")

    # this flag starts 0, is set to 1 later
    Params[12] = 0

    # kernels for subsample alignment
    Ka, Kb = getKernels(params)

    p1 = 0.95  # decay of nsp estimate in each batch

    ntot = 0
    # this keeps track of dropped templates for debugging purposes
    ndrop = np.zeros(2, dtype=np.float32, order="F")

    # this is the minimum firing rate that all templates must maintain, or be dropped
    m0 = params.minFR * params.NT / params.fs

    # allocate variables when switching to extraction phase
    # this holds spike times, clusters and other info per spike
    st3 = []  # cp.zeros((int(1e7), 5), dtype=np.float32, order='F')

    # these ones store features per spike
    # Nnearest is the number of nearest templates to store features for
    fW = LargeArrayWriter(
        ctx.path("fW", ext=".dat"), dtype=np.float32, shape=(Nnearest, -1)
    )
    # NchanNear is the number of nearest channels to take PC features from
    fWpc = LargeArrayWriter(
        ctx.path("fWpc", ext=".dat"), dtype=np.float32, shape=(NchanNear, Nrank, -1)
    )

    if params.low_memory:
        feature_path = ctx.context_path / 'spike_features'
        os.mkdir(feature_path)

    for ibatch in tqdm(range(niter), desc="Optimizing templates"):
        # korder is the index of the batch at this point in the schedule
        korder = int(irounds[ibatch])
        # k is the index of the batch in absolute terms
        logger.debug("Batch %d/%d, %d templates.", ibatch, niter, Nfilt)

        if ibatch < niter - nBatches:
            # obtained pm for this batch
            Params[8] = float(pmi[ibatch])
            pm = pmi[ibatch] * ones((Nfilt,), dtype=np.float64, order="F")

        # loading a single batch
        dataRAW = ir.data_loader.load_batch(korder)

        if ibatch == 0:
            # only on the first batch, we first get a new set of spikes from the residuals,
            # which in this case is the unmodified data because we start with no templates
            # CUDA function to get spatiotemporal clips from spike detections
            dWU, cmap = mexGetSpikes2(Params, dataRAW, wTEMP, iC)

            dWU = cp.asarray(dWU, dtype=np.float64, order="F")

            # project these into the wPCA waveforms
            dWU = cp.reshape(
                cp.dot(
                    wPCAd, cp.dot(wPCAd.T, dWU.reshape((dWU.shape[0], -1), order="F"))
                ),
                dWU.shape,
                order="F",
            )

            # initialize the low-rank decomposition with standard waves
            W = W0[:, cp.ones(dWU.shape[2], dtype=np.int32), :]
            Nfilt = W.shape[1]  # update the number of filters/templates
            # initialize the number of spikes for new templates with the minimum allowed value,
            # so it doesn't get thrown back out right away
            nsp = _extend(nsp, 0, Nfilt, m0)
            Params[1] = Nfilt  # update in the CUDA parameters

        if flag_resort:
            # this is a flag to resort the order of the templates according to best peak
            # channel
            # this is important in order to have cohesive memory requests from the GPU RAM
            # max channel (either positive or negative peak)
            iW = cp.argmax(cp.abs(dWU[nt0min - 1, :, :]), axis=0)
            # iW = int32(squeeze(iW))

            isort = cp.argsort(iW)  # sort by max abs channel
            iW = cp.asfortranarray(iW[isort])
            W = cp.asfortranarray(W[
                :, isort, :
            ])  # user ordering to resort all the other template variables
            dWU = cp.asfortranarray(dWU[:, :, isort])
            nsp = cp.asfortranarray(nsp[isort])

        # decompose dWU by svd of time and space (via covariance matrix of 61 by 61 samples)
        # this uses a "warm start" by remembering the W from the previous iteration
        W, U, mu = mexSVDsmall2(Params, dWU, W, iC, iW, Ka, Kb)

        # UtU is the gram matrix of the spatial components of the low-rank SVDs
        # it tells us which pairs of templates are likely to "interfere" with each other
        # such as when we subtract off a template
        # this needs to change (but I don't know why!)
        UtU, maskU = getMeUtU(iW, iC, mask, Nnearest, Nchan)

        # main CUDA function in the whole codebase. does the iterative template matching
        # based on the current templates, gets features for these templates if requested
        # (featW, featPC),
        # gets scores for the template fits to each spike (vexp), outputs the average of
        # waveforms assigned to each cluster (dWU0),
        # and probably a few more things I forget about
        st0, id0, x0, featW, dWU0, drez, nsp0, featPC, vexp = mexMPnu8(
            Params, dataRAW, U, W, mu, iC, iW, UtU, iList, wPCA, params
        )

        logger.debug("%d spikes.", x0.size)

        # Sometimes nsp can get transposed (think this has to do with it being
        # a single element in one iteration, to which elements are added
        # nsp, nsp0, and pm must all be row vectors (Nfilt x 1), so force nsp
        # to be a row vector.
        # nsp = cp.atleast_2d(nsp)
        # nsprow, nspcol = nsp.shape
        # if nsprow < nspcol:
        #     nsp = nsp.T
        nsp = nsp.squeeze()

        # updates the templates as a running average weighted by recency
        # since some clusters have different number of spikes, we need to apply the
        # exp(pm) factor several times, and fexp is the resulting update factor
        # for each template
        fexp = np.exp(nsp0 * cp.log(pm[:Nfilt]))
        fexp = cp.reshape(fexp, (1, 1, -1), order="F")

        # disable template updates during extraction phase
        if ibatch <= niter - nBatches - 1:
            dWU = dWU * fexp + (1 - fexp) * (
                dWU0 / cp.reshape(cp.maximum(1, nsp0), (1, 1, -1), order="F")
            )

        # nsp just gets updated according to the fixed factor p1
        nsp = nsp * p1 + (1 - p1) * nsp0

        if ibatch == niter - nBatches - 1:
            # if we reached this point, we need to disable secondary template updates
            # like dropping, and adding new templates. We need to memorize the state of the
            # templates at this timepoint, and set the processing mode to "extraction and
            # tracking"

            flag_resort = 0  # no need to resort templates by channel any more
            # flag_final = 1  # this is the "final" pass

            # final clean up, triage templates one last time
            W, U, dWU, mu, nsp, ndrop = triageTemplates2(
                params, iW, C2C, W, U, dWU, mu, nsp, ndrop
            )

            # final number of templates
            Nfilt = W.shape[1]
            Params[1] = Nfilt

            # final covariance matrix between all templates
            WtW, iList = getMeWtW(W, U, Nnearest)

            # iW is the final channel assigned to each template
            iW = cp.asfortranarray(cp.argmax(cp.abs(dWU[nt0min - 1, :, :]), axis=0))

            # extract ALL features on the last pass
            Params[
                12
            ] = 2  # this is a flag to output features (PC and template features)

            # different threshold on last pass?
            Params[2] = params.Th[
                -1
            ]  # usually the threshold is much lower on the last pass

            # memorize the state of the templates
            logger.debug("Memorized middle timepoint.")
            ir.W, ir.dWU, ir.U, ir.mu = W, dWU, U, mu
            ir.Wraw = cp.zeros(
                (U.shape[0], W.shape[0], U.shape[1]), dtype=np.float64, order="F"
            )
            for n in range(U.shape[1]):
                # temporarily use U rather Urot until I have a chance to test it
                ir.Wraw[:, :, n] = mu[n] * cp.dot(U[:, n, :], W[:, n, :].T)

            if params.low_memory:
                # Initialise array writers for the spike features of each cluster
                feature_writers = [
                    LargeArrayWriter(
                        feature_path / f'spike_features_{i}',
                        dtype=np.float32,
                        shape=(NchanNear, Nrank, -1)
                    ) for i in range(Nfilt)
                ]

        if ibatch < niter - nBatches - 1:
            # during the main "learning" phase of fitting a model
            if ibatch % 5 == 0:
                # this drops templates based on spike rates and/or similarities to
                # other templates
                W, U, dWU, mu, nsp, ndrop = triageTemplates2(
                    params, iW, C2C, W, U, dWU, mu, nsp, ndrop
                )

            Nfilt = W.shape[1]  # update the number of filters
            Params[1] = Nfilt

            # this adds new templates if they are detected in the residual
            dWU0, cmap = mexGetSpikes2(Params, drez, wTEMP, iC)

            if dWU0.shape[2] > 0:
                # new templates need to be integrated into the same format as all templates
                # apply PCA for smoothing purposes
                dWU0 = cp.reshape(
                    cp.dot(
                        wPCAd,
                        cp.dot(
                            wPCAd.T,
                            dWU0.reshape(
                                (dWU0.shape[0], dWU0.shape[1] * dWU0.shape[2]),
                                order="F",
                            ),
                        ),
                    ),
                    dWU0.shape,
                    order="F",
                )
                dWU = cp.asfortranarray(cp.concatenate((dWU, dWU0), axis=2))

                m = dWU0.shape[2]
                # initialize temporal components of waveforms
                W = cp.asfortranarray(
                    _extend(
                    W, Nfilt, Nfilt + m, W0[:, cp.ones(m, dtype=np.int32), :], axis=1
                    )
                )

                # initialize the number of spikes with the minimum allowed
                nsp = _extend(nsp, Nfilt, Nfilt + m, params.minFR * NT / params.fs)
                # initialize the amplitude of this spike with a lowish number
                mu = _extend(mu, Nfilt, Nfilt + m, 10)

                # if the number of filters exceed the maximum allowed, clip it
                Nfilt = min(params.Nfilt, W.shape[1])
                Params[1] = Nfilt

                W = W[:, :Nfilt, :]  # remove any new filters over the maximum allowed
                dWU = dWU[
                    :, :, :Nfilt
                ]  # remove any new filters over the maximum allowed
                nsp = nsp[:Nfilt]  # remove any new filters over the maximum allowed
                mu = mu[:Nfilt]  # remove any new filters over the maximum allowed

        if ibatch > niter - nBatches - 1:
            # during the final extraction pass, this keeps track of all spikes and features

            # we carefully assign the correct absolute times to spikes found in this batch
            toff = nt0min + t0 + NT*korder
            st = toff + st0

            st30 = np.c_[
                cp.asnumpy(st),  # spike times
                cp.asnumpy(id0),  # spike clusters (0-indexing)
                cp.asnumpy(x0),  # template amplitudes
                cp.asnumpy(vexp),  # residual variance of this spike
                korder * np.ones(st.size),  # batch from which this spike was found
            ]
            # Check the number of spikes.
            assert st30.shape[0] == featW.shape[1] == featPC.shape[2]

            # Sort by spike times
            ix_sort = np.argsort(st30[:,0])
            st30 = np.asfortranarray(st30[ix_sort])
            featW = cp.asfortranarray(featW[:, ix_sort])
            featPC = cp.asfortranarray(featPC[:, :, ix_sort])

            st3.append(st30)
            fW.append(featW)
            fWpc.append(featPC)

            if params.low_memory:
                # Save spike features individually for each cluster
                for i in range(Nfilt):
                    sub_array = cp.asfortranarray(featPC[:, :, id0 == i])
                    feature_writers[i].append(sub_array)


            ntot = ntot + x0.size  # keeps track of total number of spikes so far

        if ibatch % 100 == 0:
            # this is some of the relevant diagnostic information to be printed during training
            # logger.info(
            #     (
            #         "%d / %d batches, %d units, nspks: %2.4f, mu: %2.4f, "
            #         "nst0: %d, merges: %2.4f, %2.4f"
            #     ),
            #     ibatch,
            #     niter,
            #     Nfilt,
            #     nsp.sum(),
            #     median(mu),
            #     st0.size,
            #     *ndrop
            # )

            if sanity_plots:
                assert plot_widgets is not None, "if sanity_plots is set, then plot_widgets cannot be None"
                plot_diagnostics(W, U, mu, nsp, plot_widgets[plot_pos])

        free_gpu_memory()

    # Close the large array writers and save the JSON metadata files to disk.
    fW.close()
    fWpc.close()

    if params.low_memory:
        # Close the large array writers for each cluster's spike features
        for writer in feature_writers:
            writer.close()

    # just display the total number of spikes
    logger.info("Found %d spikes.", ntot)

    # Save results to the ctx.intermediate object.
    ir.st3 = np.concatenate(st3, axis=0)

    # the similarity score between templates is simply the correlation,
    # taken as the max over several consecutive time delays
    ir.simScore = cp.asnumpy(cp.max(WtW, axis=2))

    # NOTE: these are now already saved by LargeArrayWriter
    # fWa = np.concatenate(fW, axis=-1)
    # fWpca = np.concatenate(fWpc, axis=-1)

    # the template features are stored in cProj, like in Kilosort1
    # ir.cProj = fWa.T
    # the neihboring templates idnices are stored in iNeigh
    ir.iNeigh = cp.asnumpy(iList)

    #  permute the PC projections in the right order
    # ir.cProjPC = np.transpose(fWpca, (2, 1, 0))
    # iNeighPC keeps the indices of the channels corresponding to the PC features
    ir.iNeighPC = cp.asnumpy(iC[:, iW])

    # Number of spikes.
    assert ir.st3.shape[0] == fW.shape[-1] == fWpc.shape[-1]

    # Save cluster times and IDs at this stage if requested
    if params.save_temp_files:
        np.save(ctx.context_path / 'temp_splits' / 'st3_learn.npy', ir.st3)

    # # this whole next block is just done to compress the compressed templates
    # # we separately svd the time components of each template, and the spatial components
    # # this also requires a careful decompression function, available somewhere in the GUI code
    # nKeep = min(Nchan * 3, 20)  # how many PCs to keep
    # W_a = np.zeros((nt0 * Nrank, nKeep, Nfilt), dtype=np.float32)
    # W_b = np.zeros((nBatches, nKeep, Nfilt), dtype=np.float32)
    # U_a = np.zeros((Nchan * Nrank, nKeep, Nfilt), dtype=np.float32)
    # U_b = np.zeros((nBatches, nKeep, Nfilt), dtype=np.float32)
    #
    # for j in tqdm(range(Nfilt), desc="Compressing templates"):
    #     # do this for every template separately
    #     WA = np.reshape(ir.WA[:, j, ...], (-1, nBatches), order="F")
    #     # svd on the GPU was faster for this, but the Python randomized CPU version
    #     # might be faster still
    #     # WA = gpuArray(WA)
    #     A, B, C = svdecon_cpu(WA)
    #     # W_a times W_b results in a reconstruction of the time components
    #     W_a[:, :, j] = np.dot(A[:, :nKeep], B[:nKeep, :nKeep])
    #     W_b[:, :, j] = C[:, :nKeep]
    #
    #     UA = np.reshape(ir.UA[:, j, ...], (-1, nBatches), order="F")
    #     # UA = gpuArray(UA)
    #     A, B, C = svdecon_cpu(UA)
    #     # U_a times U_b results in a reconstruction of the time components
    #     U_a[:, :, j] = np.dot(A[:, :nKeep], B[:nKeep, :nKeep])
    #     U_b[:, :, j] = C[:, :nKeep]
    #
    # logger.info("Finished compressing time-varying templates.")

    #TODO
    # Keep st3, wPCA, wTEMP, simScore, iNeigh, iNeighPC, W, U, dWU
    #

    return Bunch(
        wPCA=wPCA[:, :Nrank],
        wTEMP=wTEMP,
        st3=ir.st3,
        simScore=ir.simScore,
        # cProj=ir.cProj,
        # cProjPC=ir.cProjPC,
        iNeigh=ir.iNeigh,
        iNeighPC=ir.iNeighPC,
        W=ir.W,
        U=ir.U,
        dWU=ir.dWU,
        mu=ir.mu,
        # W_a=W_a,
        # W_b=W_b,
        # U_a=U_a,
        # U_b=U_b,
    )

def compress_templates(ctx):
    """
    Individually compress the temporal and spatial components of the spike templates
    """
    Nbatch = ctx.intermediate.Nbatch
    params = ctx.params
    probe = ctx.probe
    ir = ctx.intermediate

    Nchan = probe.Nchan
    nt0 = params.nt0
    Nfilt = ir.UA.shape[1]
    Nbatch = ir.Nbatch

    Nrank = 3

    nKeep = min(Nchan * 3, 20)  # how many PCs to keep
    W_a = np.zeros((nt0 * Nrank, nKeep, Nfilt), dtype=np.float32)
    W_b = np.zeros((Nbatch, nKeep, Nfilt), dtype=np.float32)
    U_a = np.zeros((Nchan * Nrank, nKeep, Nfilt), dtype=np.float32)
    U_b = np.zeros((Nbatch, nKeep, Nfilt), dtype=np.float32)

    for j in tqdm(range(Nfilt), desc="Compressing templates"):
        # do this for every template separately
        WA = np.reshape(ir.WA[:, j, ...], (-1, Nbatch), order="F")
        # svd on the GPU was faster for this, but the Python randomized CPU version
        # might be faster still
        # WA = gpuArray(WA)
        A, B, C = svdecon_cpu(WA)
        # W_a times W_b results in a reconstruction of the time components
        W_a[:, :, j] = np.dot(A[:, :nKeep], B[:nKeep, :nKeep])
        W_b[:, :, j] = C[:, :nKeep]

        UA = np.reshape(ir.UA[:, j, ...], (-1, Nbatch), order="F")
        # UA = gpuArray(UA)
        try:
            A, B, C = svdecon_cpu(UA)
        except np.linalg.LinAlgError:
            logger.error("SVD did not converge")
            continue
        # U_a times U_b results in a reconstruction of the time components
        U_a[:, :, j] = np.dot(A[:, :nKeep], B[:nKeep, :nKeep])
        U_b[:, :, j] = C[:, :nKeep]

    logger.info("Finished compressing time-varying templates.")

    #TODO
    # Write these to the disk

    return Bunch(
        W_a=W_a,
        W_b=W_b,
        U_a=U_a,
        U_b=U_b,
    )



import numpy as np

from pykilosort.postprocess import ccg


def slow_ccg(st1, st2, nbins, tbin):
    diffs = np.rint((st1 - st2.reshape(-1,1)).flatten() / tbin).astype('int')
    output = np.histogram(diffs, bins=np.arange(-nbins, nbins+2))[0]
    return output


def test_ccg():

    spike_times = np.cumsum(np.random.randint(100, 30000, 500)) / 30000
    nbins = 500
    tbin = 0.1

    output = ccg(spike_times, spike_times, nbins, tbin)
    output_comparison = slow_ccg(spike_times, spike_times, nbins, tbin)

    assert np.allclose(output[1:-1], output_comparison[1:-1])
